"""Module for converting T1 of ADNI."""

from pathlib import Path
from typing import Any, Iterable, Optional

import pandas as pd

__all__ = ["convert_t1"]


def convert_t1(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert T1 MR images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    destination_dir : Path
        The path to the destination BIDS directory.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        (
            f"Calculating paths of {ADNIModalityConverter.T1.value} images. "
            f"Output will be stored in {conversion_dir}."
        ),
        lvl="info",
    )
    images = _compute_t1_paths(source_dir, csv_dir, subjects, conversion_dir)
    cprint(
        f"Paths of {ADNIModalityConverter.T1.value} images found. Exporting images into BIDS ...",
        lvl="info",
    )
    paths_to_bids(
        images,
        destination_dir,
        ADNIModalityConverter.T1,
        mod_to_update=mod_to_update,
        n_procs=n_procs,
    )
    cprint(msg=f"{ADNIModalityConverter.T1.value} conversion done.", lvl="debug")


def _compute_t1_paths(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
) -> pd.DataFrame:
    """Compute the paths to T1 MR images and store them in a TSV file.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    subjects : list of str
        The subjects list.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    Returns
    -------
    images : pd.DataFrame
        A dataframe with all the paths to the T1 MR images that will be converted into BIDS.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv
    from clinica.utils.stream import cprint

    from ._image_path_utils import find_image_path
    from ._visits_utils import visits_to_timepoints

    t1_dfs_list = []
    t1_df = _initialize_t1_df()
    adni_merge = load_clinical_csv(csv_dir, "ADNIMERGE")
    mprage_meta = load_clinical_csv(csv_dir, "MPRAGEMETA")
    mri_quality = load_clinical_csv(csv_dir, "MRIQUALITY")
    mayo_mri_qc = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15")

    # Keep only T1 scans
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "T1"]

    # We will convert the images for each subject in the subject list
    for subject in subjects:
        # Filter ADNIMERGE, MPRAGE METADATA and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subject]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mprage_meta_subj = mprage_meta[mprage_meta.SubjectID == subject]
        mprage_meta_subj = mprage_meta_subj.sort_values("ScanDate")

        mri_quality_subj = mri_quality[mri_quality.RID == int(subject[-4:])]
        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subject[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(
            subject=subject,
            mri_list_subj=mprage_meta_subj,
            adnimerge_subj=adnimerge_subj,
            modality="T1",
            visit_field="Visit",
            scandate_field="ScanDate",
        )
        for visit_info in visits.keys():
            cohort = visit_info[1]
            timepoint = visit_info[0]
            visit_str = visits[visit_info]
            selected_scan = None

            if cohort in ("ADNI1", "ADNIGO", "ADNI2"):
                selected_scan = _select_preferred_scan_for_subject_in_adni1go2(
                    subject_id=subject,
                    timepoint=timepoint,
                    visit_str=visit_str,
                    mprage_meta_subj=mprage_meta_subj,
                    mri_quality_subj=mri_quality_subj,
                    mayo_mri_qc_subj=mayo_mri_qc_subj,
                    preferred_field_strength=1.5 if cohort == "ADNI1" else 3.0,
                )
            elif cohort == "ADNI3":
                selected_scan = _select_preferred_scan_for_subject_in_adni3(
                    subject_id=subject,
                    timepoint=timepoint,
                    visit_str=visit_str,
                    mprage_meta_subj=mprage_meta_subj,
                    mayo_mri_qc_subj=mayo_mri_qc_subj,
                )
            else:
                cprint(
                    f"Subject {subject} visit {visit_str} belongs to an unknown cohort: {cohort}"
                )
            if selected_scan is not None:
                t1_dfs_list.append(pd.DataFrame(selected_scan, index=["i"]))
    if t1_dfs_list:
        t1_df = pd.concat(t1_dfs_list, ignore_index=True)

    # Removing known exceptions from images to convert
    if not t1_df.empty:
        error_indices = t1_df.index[
            t1_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in _get_known_conversion_errors()),
                axis=1,
            )
        ]
        t1_df.drop(error_indices, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(t1_df, source_dir, modality="T1")
    images.to_csv(conversion_dir / "t1_paths.tsv", sep="\t", index=False)

    return images


def _initialize_t1_df() -> pd.DataFrame:
    return pd.DataFrame(
        columns=[
            "Subject_ID",
            "VISCODE",
            "Visit",
            "Sequence",
            "Scan_Date",
            "Study_ID",
            "Field_Strength",
            "Series_ID",
            "Image_ID",
            "Original",
        ]
    )


def _get_known_conversion_errors() -> Iterable[tuple[str, str]]:
    return [  # Eq_1
        ("031_S_0830", "m48"),
        ("100_S_0995", "m18"),
        ("031_S_0867", "m48"),
        ("100_S_0892", "m18"),
        ("003_S_6264", "m12"),
        ("011_S_4105", "m72"),
        # Empty folders
        # ('029_S_0845', 'm24'),
        # ('094_S_1267', 'm24'),
        # ('029_S_0843', 'm24'),
        # ('027_S_0307', 'm48'),
        # ('057_S_1269', 'm24'),
        # ('036_S_4899', 'm03'),
        # ('033_S_1016', 'm120'),
        # ('130_S_4984', 'm12'),
        # ('027_S_4802', 'm06'),
        # ('131_S_0409', 'bl'),
        # ('082_S_4224', 'm24'),
        # ('006_S_4960', 'bl'),
        # ('006_S_4960', 'm03'),
        # ('006_S_4960', 'm06'),
        # ('006_S_4960', 'm12'),
        # ('006_S_4960', 'm24'),
        # ('006_S_4960', 'm36'),
        # ('006_S_4960', 'm72'),
        # ('022_S_5004', 'bl'),
        # ('022_S_5004', 'm03'),
        # Several images: T1wa ...
        ("006_S_4485", "m84"),
        ("029_S_2395", "m72"),
        ("114_S_6039", "bl"),
        ("016_S_4952", "m48"),
    ]


def _select_preferred_scan_for_subject_in_adni1go2(
    subject_id: str,
    timepoint: str,
    visit_str: str,
    mprage_meta_subj: pd.DataFrame,
    mri_quality_subj: pd.DataFrame,
    mayo_mri_qc_subj: pd.DataFrame,
    preferred_field_strength: float = 3.0,
) -> Optional[dict[str, Any]]:
    """Select the preferred scan for a subject in a visit, given the subject belongs to ADNI 1, Go or 2 cohorts.

    Parameters
    ----------
    subject_id : str
        A string containing subject ID.

    timepoint : str
        A string of visit code in months.

    visit_str : str
        A string of visit name.

    mprage_meta_subj : pd.DataFrame
        A DataFrame of MPRAGE metadata of images corresponding to the subject.

    mri_quality_subj : pd.DataFrame
        A DatFrame of MR image quality of images corresponding to the subject.

    mayo_mri_qc_subj : pd.DatFrame
        A DataFrame of MAYO Clinic MR image quality of images corresponding to the subject.

    preferred_field_strength : float, optional
        Field strength that is preferred in case there are several image acquisitions.

    Returns
    -------
    dict or None:
        Dictionary containing selected scan information.
        Returns None if no scan found.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars

    # filter out images that do not pass QC
    mprage_meta_subj = mprage_meta_subj[
        mprage_meta_subj.apply(
            lambda x: _check_qc(x, subject_id, visit_str, mri_quality_subj), axis=1
        )
    ]
    filtered_mprage = _get_last_processed_scan_for_visit(mprage_meta_subj, visit_str)

    # If no N3 processed image found (it means there are no processed images at all), get best original image
    if filtered_mprage.empty:
        return _select_best_original_image(
            subject_id,
            timepoint,
            visit_str,
            mprage_meta_subj,
            mayo_mri_qc_subj,
            preferred_field_strength,
        )

    # If there are images with different magnetic field strength, prefer 1.5T images for ADNI1, 3.0T otherwise
    if len(filtered_mprage.MagStrength.unique()) > 1:
        filtered_mprage = filtered_mprage[
            filtered_mprage.MagStrength == preferred_field_strength
        ]

    # Sort by Series ID in case there are several images, so we keep the one acquired first
    filtered_mprage = filtered_mprage.sort_values("SeriesID")
    scan = filtered_mprage.iloc[0]

    n3 = scan.Sequence.find("N3")
    # Sequence ends in 'N3' or in 'N3m'
    sequence = scan.Sequence[: n3 + 2 + int(scan.Sequence[n3 + 2] == "m")]
    sequence = replace_sequence_chars(sequence)

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": sequence,
        "Scan_Date": scan.ScanDate,
        "Study_ID": str(scan.StudyID),
        "Series_ID": str(scan.SeriesID),
        "Image_ID": str(scan.ImageUID),
        "Field_Strength": scan.MagStrength,
        "Original": False,
    }


def _get_last_processed_scan_for_visit(
    mprage_meta_subj: pd.DataFrame,
    visit_str: str,
    unwanted_series_id: Optional[tuple[str]] = None,
) -> pd.DataFrame:
    """Filter content of Dataframe containing MPRAGE metadata to obtain last processed image for the specified visit.

    Parameters
    ----------
    mprage_meta_subj : pd.DataFrame
        A DataFrame of MPRAGE metadata of images corresponding to the subject.

    visit_str : str
        The visit name.

    unwanted_series_id : tuple or str, optional
        Series ID of scans not passing QC.

    Returns
    -------
    pd.Dataframe
    """
    unwanted_series_id = unwanted_series_id or ()
    mprage_meta_subj = mprage_meta_subj[
        ~mprage_meta_subj.SeriesID.isin(unwanted_series_id)
    ]
    # Get the preferred scan (image series that has been Scaled)
    filtered_mprage = mprage_meta_subj[
        (mprage_meta_subj["Orig/Proc"] == "Processed")
        & (mprage_meta_subj.Visit == visit_str)
        & mprage_meta_subj.Sequence.str.endswith("Scaled", na=False)
    ]
    # If no preferred image found, get N3 processed image (N3m)
    if filtered_mprage.empty:
        filtered_mprage = mprage_meta_subj[
            (mprage_meta_subj["Orig/Proc"] == "Processed")
            & (mprage_meta_subj.Visit == visit_str)
            & mprage_meta_subj.Sequence.str.endswith("N3m", na=False)
        ]

    return filtered_mprage


def _select_preferred_scan_for_subject_in_adni3(
    subject_id: str,
    timepoint: str,
    visit_str: str,
    mprage_meta_subj: pd.DataFrame,
    mayo_mri_qc_subj: pd.DataFrame,
) -> Optional[dict[str, Any]]:
    """Select the preferred scan for a subject in a visit, given the subject belongs to ADNI 3 cohort.

    Parameters
    ----------
    subject_id : str
        A string containing subject ID.

    timepoint : str
        A string of visit code in months.

    visit_str : str
        A string of visit name.

    mprage_meta_subj : pd.DataFrame
        A DataFrame of MPRAGE metadata of images corresponding to the subject.

    mayo_mri_qc_subj : pd.DataFrame
        A DatFrame of MAYO Clinic MR image quality of images corresponding to the subject.

    Returns
    -------
    dict or None :
        Dictionary containing selected scan information.
        Returns None if no scan found.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    filtered_scan = mprage_meta_subj[
        (mprage_meta_subj["Orig/Proc"] == "Original")
        & (mprage_meta_subj.Visit == visit_str)
        & mprage_meta_subj.Sequence.str.contains("accel", case=False, na=False)
        & ~mprage_meta_subj.Sequence.str.lower().str.endswith("_nd", na=False)
    ]
    if filtered_scan.empty:
        cprint(
            f"NO MPRAGE Meta for ADNI3: {subject_id} for visit {timepoint} - {visit_str}"
        )
        return None
    scan = _select_scan_from_qc(
        filtered_scan, mayo_mri_qc_subj, preferred_field_strength=3.0
    )
    sequence = replace_sequence_chars(scan.Sequence)

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": sequence,
        "Scan_Date": scan.ScanDate,
        "Study_ID": str(scan.StudyID),
        "Series_ID": str(scan.SeriesID),
        "Image_ID": str(scan.ImageUID),
        "Field_Strength": scan.MagStrength,
        "Original": True,
    }


def _select_best_original_image(
    subject_id: str,
    timepoint: str,
    visit_str: str,
    mprage_meta_subj: pd.DataFrame,
    mayo_mri_qc_subj: pd.DataFrame,
    preferred_field_strength: float = 3.0,
) -> Optional[dict[str, Any]]:
    """Select the preferred scan for a subject in a visit, given the scan is not preprocessed.

    Parameters
    ----------
    subject_id : str
        A string containing subject ID.

    timepoint : str
        A string of visit code in months.

    visit_str : str
        A string of visit name.

    mprage_meta_subj : pd.DataFrame
        A DataFrame of MPRAGE metadata of images corresponding to the subject.

    mayo_mri_qc_subj : pd.DataFrame
        A DatFrame of MAYO Clinic MR image quality of images corresponding to the subject.

    preferred_field_strength : float, optional
        Field strength that is preferred in case there are several image acquisitions.

    Returns
    -------
    dict or None :
        Dictionary containing selected scan information.
        Returns None if no scan found.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    mprage_meta_subj_orig = mprage_meta_subj[
        mprage_meta_subj["Orig/Proc"] == "Original"
    ]
    cond_mprage = (
        (mprage_meta_subj_orig.Visit == visit_str)
        & mprage_meta_subj_orig.Sequence.str.contains(
            "mprage|mp-rage|mp rage", case=False, na=False
        )
        & ~mprage_meta_subj_orig.Sequence.str.contains("2", na=False)
    )
    cond_spgr = (
        (mprage_meta_subj_orig.Visit == visit_str)
        & mprage_meta_subj_orig.Sequence.str.contains("spgr", case=False, na=False)
        & ~mprage_meta_subj_orig.Sequence.str.contains("acc", case=False, na=False)
    )
    filtered_scan = mprage_meta_subj_orig[cond_mprage | cond_spgr]

    if filtered_scan.empty:
        cprint(f"NO MPRAGE Meta: {subject_id} for visit {timepoint} - {visit_str}")
        return None
    scan = _select_scan_from_qc(
        filtered_scan, mayo_mri_qc_subj, preferred_field_strength
    )
    sequence = replace_sequence_chars(scan.Sequence)

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": sequence,
        "Scan_Date": scan.ScanDate,
        "Study_ID": str(scan.StudyID),
        "Series_ID": str(scan.SeriesID),
        "Image_ID": str(scan.ImageUID),
        "Field_Strength": scan.MagStrength,
        "Original": True,
    }


def _select_scan_from_qc(
    scans_meta: pd.DataFrame,
    mayo_mri_qc_subj: pd.DataFrame,
    preferred_field_strength: float,
) -> pd.DataFrame:
    """Select a scan from a list of scans taking into account available QC.

    Parameters
    ----------
    scans_meta : pd.DataFrame
        A DataFrame containing the metadata for images to choose among.

    mayo_mri_qc_subj : pd.DataFrame
        A DatFrame of MAYO Clinic MR image quality of images corresponding to the subject.

    preferred_field_strength : pd.DataFrame
        Field strength that is preferred in case there are several image acquisitions.

    Returns
    -------
    pd.DataFrame :
        A DataFrame row containing selected scan.
    """
    import numpy as np

    multiple_mag_strength = len(scans_meta.MagStrength.unique()) > 1

    # Select preferred_field_strength images
    if multiple_mag_strength:
        # Save for later the scans with the non preferred magnetic field strength
        not_preferred_scan = scans_meta[
            scans_meta.MagStrength != preferred_field_strength
        ]
        # Filtering to keep only the scans with the preferred magnetic field strength
        scans_meta = scans_meta[scans_meta.MagStrength == preferred_field_strength]
    else:
        not_preferred_scan = None

    if scans_meta.MagStrength.unique()[0] == 3.0:
        id_list = scans_meta.ImageUID.unique()
        image_ids = ["I" + str(imageuid) for imageuid in id_list]
        int_ids = [int(imageuid) for imageuid in id_list]
        images_qc = mayo_mri_qc_subj[mayo_mri_qc_subj.loni_image.isin(image_ids)]

        if not images_qc.empty:
            selected_image = None
            # Check if there is only one selected series image.
            if np.sum(images_qc.series_selected) == 1:
                selected_image = (
                    images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
                )
            # Otherwise, select the one with the best QC if available.
            else:
                images_not_rejected = images_qc[images_qc.series_quality < 4]

                if images_not_rejected.empty:
                    # There are no images that passed the qc,
                    # so we'll try to see if there are other images without qc.
                    # Otherwise, return None.
                    qc_ids = set(
                        [int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()]
                    )
                    no_qc_ids = list(set(int_ids) - qc_ids)

                    # If none of images passed qc the scan is None, otherwise:
                    if no_qc_ids:
                        no_qc_scans_meta = scans_meta[
                            scans_meta.ImageUID.isin(no_qc_ids)
                        ]
                        return _select_scan_no_qc(no_qc_scans_meta)
                else:
                    # We select the image with the best (lower) QC.
                    # If no positive QC available we choose image with -1 (not performed)
                    series_quality = [
                        q if q > 0 else 4
                        for q in list(images_not_rejected.series_quality)
                    ]
                    best_q = np.amin(series_quality)
                    if best_q == 4:
                        best_q = -1
                    images_best_qc = images_not_rejected[
                        images_not_rejected.series_quality == best_q
                    ]
                    if len(images_best_qc) == 1:
                        selected_image = images_best_qc.iloc[0].loni_image[1:]
                    else:
                        best_ids = [
                            int(x[1:]) for x in images_best_qc.loni_image.unique()
                        ]
                        best_qc_meta = scans_meta[scans_meta.ImageUID.isin(best_ids)]
                        return _select_scan_no_qc(best_qc_meta)

            # If we did not find an image passing QC for the preferred magnetic field strength,
            # then we will choose between the other available images with other magnetic field strength
            if not selected_image and multiple_mag_strength:
                scans_meta = not_preferred_scan
            else:
                scan = scans_meta[scans_meta.ImageUID == int(selected_image)].iloc[0]
                return scan

    # 1.5T (no available QC)
    return _select_scan_no_qc(scans_meta)


def _select_scan_no_qc(scans_meta: pd.DataFrame) -> pd.DataFrame:
    """Select a scan from available scans in case there is not an available QC.

    Parameters
    ----------
    scans_meta : pd.DataFrame
        A DataFrame containing the metadata for images to choose among.

    Returns
    -------
    pd.DataFrame :
        DataFrame row containing selected scan.
    """
    # We choose the first scan (not containing repeat in name)
    selected_scan = scans_meta[
        ~scans_meta.Sequence.str.contains("repeat", case=False, na=False)
    ]
    if selected_scan.empty:
        selected_scan = scans_meta
    scan = selected_scan.iloc[0]
    return scan


def _check_qc(
    scan: pd.DataFrame, subject_id: str, visit_str: str, mri_quality_subj: pd.DataFrame
) -> bool:
    """Check if a scan has passed check quality, if it was performed.

    Parameters
    ----------
    scan : pd.DataFrame
        A DataFrame row containing scan information.

    subject_id : str
        A string containing subject ID.

    visit_str : str
        A string of visit name.

    mri_quality_subj : pd.DataFrame
        A DatFrame of MR image quality of images corresponding to the subject.

    Returns
    -------
    bool :
        Boolean, True if image passed QC or if there is no available QC, False otherwise.
    """
    from clinica.utils.stream import cprint

    # Check if QC exists for image series
    qc = mri_quality_subj[mri_quality_subj.LONIUID == "S" + str(scan.SeriesID)]

    # If QC exists and failed we keep the other scan (in case 2 scans were performed)
    if not qc.empty and qc.iloc[0].PASS != 1:
        cprint("QC found but NOT passed", lvl="info")
        cprint(
            msg=(
                f"Subject {subject_id} for visit {visit_str} "
                f"- Series: {str(scan.SeriesID)} - Study: {str(scan.StudyID)}"
            ),
            lvl="info",
        )
        return False
    return True



from typing import Optional, Sequence

import numpy as np
import pandas as pd

__all__ = ["select_image_qc"]


def select_image_qc(id_list: Sequence[str], mri_qc_subj: pd.DataFrame) -> Optional[int]:
    """Select image from several scans according to QC.

    Args:
        id_list: List of images identifiers to choose among
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: int - Chosen image identifier
    """
    if len(id_list) == 0:
        return None

    image_ids = ["I" + str(imageuid) for imageuid in id_list]
    int_ids = [int(imageuid) for imageuid in id_list]
    images_qc = mri_qc_subj[mri_qc_subj.loni_image.isin(image_ids)]

    if images_qc.empty:
        return max(int_ids)

    if np.sum(images_qc.series_selected) == 1:
        selected_image = (
            images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
        )
    else:
        images_not_rejected = images_qc[images_qc.series_quality < 4]

        if images_not_rejected.empty:
            # There are no images that passed the qc,
            # so we'll try to see if there are other images without qc.
            # Otherwise, return None.
            qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
            no_qc_ids = list(set(int_ids) - qc_ids)

            if len(no_qc_ids) == 0:
                return None
            else:
                return max(no_qc_ids)

        # We select the image with the best (lower) QC.
        # If no positive QC available we choose image with -1 (not performed)
        series_quality = [
            q if q > 0 else 4 for q in list(images_not_rejected.series_quality)
        ]
        best_q = np.amin(series_quality)
        if best_q == 4:
            best_q = -1
        images_best_qc = images_not_rejected[
            images_not_rejected.series_quality == best_q
        ]
        if len(images_best_qc) == 1:
            selected_image = images_best_qc.iloc[0].loni_image[1:]
        else:
            best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
            selected_image = max(best_ids)

    return int(selected_image)



import pydra
from pydra import Workflow

from clinica.pydra.engine import clinica_io


def _check_pipeline_parameters(parameters: dict) -> dict:
    """Check the parameters passed to the pipeline.

    Parameters
    ----------
    parameters : dict
        Dictionary of parameters to analyze.

    Returns
    -------
    dict :
        Cleaned dictionary of parameters.
    """
    from clinica.utils.exceptions import ClinicaException

    # PET-Volume pipeline
    if parameters["orig_input_data_ml"] == "pet-volume":
        for param in ("acq_label", "suvr_reference_region"):
            if not parameters[param]:
                raise ClinicaException(
                    f"You selected pet-volume pipeline without setting --{param} flag. "
                    "Clinica will now exit."
                )
    return parameters


def _build_query(parameters: dict) -> dict:
    if parameters["orig_input_data_ml"] == "t1-volume":
        return {
            "modulation": True,
            "tissue_number": 1,
            "group_label": parameters["group_label"],
        }
    if parameters["orig_input_data_ml"] == "pet-volume":
        return {
            "acq_label": parameters["acq_label"],
            "suvr_reference_region": parameters["suvr_reference_region"],
            "use_brainmasked_image": False,
            "use_pvc_data": parameters["use_pvc_data"],
            "fwhm": 0,
            "group_label": parameters["group_label"],
        }


@clinica_io
def build_core_workflow(name: str = "core", parameters={}) -> Workflow:
    """Build the core workflow for the machine learning spatial svm pipeline.

    Parameters
    ----------
    name : str, optional
        The name of the workflow. Default="core".

    parameters : dict, optional
        Dictionary of parameters to be used
        within the workflow.
        Default={}.

    Returns
    -------
    wf : Workflow
        The core
    """
    from typing import Any

    import clinica.pydra.machine_learning_spatial_svm.tasks as utils

    parameters = _check_pipeline_parameters(parameters)
    query = _build_query(parameters)
    input_name = parameters["orig_input_data_ml"].replace("-", "_")
    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", Any),
            (
                input_name,
                dict,
                query,
                {"mandatory": True},
            ),
            (
                "dartel_template",
                dict,
                {"group_label": parameters["group_label"]},
                {"mandatory": True},
            ),
        ],
        bases=(pydra.specs.BaseSpec,),
    )
    wf = Workflow(name, input_spec=input_spec)

    wf.split((input_name))

    wf.add(
        utils.obtain_g_fisher_tensor_task(
            name="obtain_g_fisher_tensor_task",
            dartel_input=wf.lzin.dartel_template,
            FWHM=parameters["fwhm"],
        )
    )
    wf.add(
        utils.obtain_time_step_estimation_task(
            name="obtain_time_step_estimation_task",
            dartel_input=wf.lzin.dartel_template,
            g=wf.obtain_g_fisher_tensor_task.lzout.g_fisher_tensor,
            FWHM=parameters["fwhm"],
        )
    )
    wf.add(
        utils.heat_solver_equation_task(
            name="heat_solver_equation_task",
            input_image=getattr(wf.lzin, input_name),
            g=wf.obtain_g_fisher_tensor_task.lzout.g_fisher_tensor,
            FWHM=parameters["fwhm"],
            t_step=wf.obtain_time_step_estimation_task.lzout.t_step,
            dartel_input=wf.lzin.dartel_template,
        )
    )
    wf.set_output(
        [
            (
                "fisher_tensor_path",
                wf.obtain_g_fisher_tensor_task.lzout.output_fisher_tensor,
            ),
            ("json_file", wf.obtain_time_step_estimation_task.lzout.output_data),
            ("regularized_image", wf.heat_solver_equation_task.lzout.regularized),
        ]
    )
    return wf



from collections import namedtuple
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from clinica.utils.image import HemiSphere

__all__ = [
    "ImageID",
    "extract_image_id_from_freesurfer_id",
    "generate_regional_measures",
]


ImageID = namedtuple("ImageID", ["participant_id", "session_id", "long_id"])


class InfoType(str, Enum):
    VOLUME = "volume"
    THICKNESS = "thickness"
    AREA = "area"
    MEANCURV = "meancurv"


class ColumnType(str, Enum):
    PARCELLATION = "parcellation"
    SEGMENTATION = "segmentation"


def generate_regional_measures(
    segmentation_path: Union[str, Path],
    subject_id: str,
    atlases: Optional[list[str]] = None,
    output_dir: Optional[Path] = None,
) -> None:
    """Read stats files and generate regional measures TSV files.

    The stats files are located in
    <segmentation_path>/<subject_id>/stats/*.stats
    and the generated TSV files will be located in the folder
    <segmentation_path>/regional_measures

    Notes
    -----
    The .stats files contain:
        - a table with statistical information (e.g., structure volume)
        - secondary statistical information with all lines starting
          with the sentence '# Measure'.

    The .tsv files return the relevant statistical information from both sources.

    Parameters
    ----------
    segmentation_path : str
        Path to the FreeSurfer segmentation.

    subject_id : str
        Subject ID in the form `sub-CLNC01_ses-M00`, `sub-CLNC01_long-M00M18` or
        `sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18`.

    atlases : List[str], optional
        The atlas names to be used.
        By default, the following atlases will be used:
            - "desikan"
            - "destrieux"
            - "ba"

    output_dir : str, optional
        Path to the folder where the .tsv stats files will be stored.
        If no output folder is provided, files will be written to
        [path_segmentation]/regional_measures.
    """
    segmentation_path = Path(segmentation_path)
    atlases = atlases or ["desikan", "destrieux", "ba"]
    if not isinstance(atlases, list):
        raise ValueError(
            f"atlases should be a list of strings. {type(atlases)} was provided instead."
        )
    prefix = _get_prefix(subject_id)
    seg_path = segmentation_path.expanduser()
    stats_folder = seg_path / subject_id / "stats"

    if not stats_folder.is_dir():
        raise FileNotFoundError(
            f"Image {prefix.replace('_', ' | ')} does not contain FreeSurfer segmentation"
        )
    output_dir = output_dir or seg_path / "regional_measures"
    output_dir.mkdir(exist_ok=True)
    _generate_tsv_for_parcellation(stats_folder, output_dir, prefix, atlases)
    _generate_tsv_for_segmentation(stats_folder, output_dir, prefix)


def _get_prefix(subject_id: str) -> str:
    image_id = extract_image_id_from_freesurfer_id(subject_id)

    return "_".join([i for i in image_id if i])


def extract_image_id_from_freesurfer_id(freesurfer_id: str) -> ImageID:
    """Extract image ID from longitudinal segmentation folder.

    This function will extract participant, session and longitudinal ID from `freesurfer_id`.

    Parameters
    ----------
    freesurfer_id : str
        The freesurfer ID.

    Returns
    -------
    image_id: ImageID
        NamedTuple containing the following fields:
            - "participant_id": The extracted participant ID.
            - "session_id": The extracted session ID.
              Note that this field can be an empty string.
            - "long_id": The extracted longitudinal ID.
              Note that this field can be an empty string.

    Examples
    --------
    >>> from clinica.pipelines.anatomical.freesurfer.utils import extract_image_id_from_freesurfer_id
    >>> extract_image_id_from_freesurfer_id('sub-CLNC001_ses-M000')
    ImageID(participant_id='sub-CLNC001', session_id='ses-M000', long_id='')
    >>> extract_image_id_from_freesurfer_id('sub-CLNC001_long-M018')
    ImageID(participant_id='sub-CLNC001', session_id='', long_id='long-M018')
    >>> extract_image_id_from_freesurfer_id('sub-CLNC001_ses-M000.long.sub-CLNC001_long-M000M018')
    ImageID(participant_id='sub-CLNC001', session_id='ses-M000', long_id='long-M000M018')
    """
    if ".long." in freesurfer_id:
        return _extract_image_id_from_longitudinal_id_dot(freesurfer_id)
    if "long-" in freesurfer_id:
        return _extract_image_id_from_longitudinal_id_dash(freesurfer_id)
    try:
        return _extract_image_id_from_cross_sectional_id(freesurfer_id)
    except IndexError:
        raise ValueError(
            f"The provided Freesurfer ID {freesurfer_id} could not be parsed."
        )


def _extract_image_id_from_longitudinal_id_dot(freesurfer_id: str) -> ImageID:
    """Handle cases like 'sub-CLNC001_ses-M000.long.sub-CLNC001_long-M000M018'."""
    participant_id = freesurfer_id.split(".long.")[0].split("_")[0]
    session_id = freesurfer_id.split(".long.")[0].split("_")[1]
    long_id = freesurfer_id.split(".long.")[1].split("_")[1]

    return ImageID(participant_id, session_id, long_id)


def _extract_image_id_from_longitudinal_id_dash(freesurfer_id: str) -> ImageID:
    """Handle cases like 'sub-CLNC001_long-M000M018'."""
    participant_id = freesurfer_id.split("_")[0]
    session_id = ""
    long_id = freesurfer_id.split("_")[1]

    return ImageID(participant_id, session_id, long_id)


def _extract_image_id_from_cross_sectional_id(freesurfer_id: str) -> ImageID:
    """Handle cases like 'sub-CLNC001_ses-M000'."""
    participant_id = freesurfer_id.split("_")[0]
    session_id = freesurfer_id.split("_")[1]
    long_id = ""

    return ImageID(participant_id, session_id, long_id)


def _generate_tsv_for_parcellation(
    stats_folder: Path,
    output_dir: Path,
    prefix: str,
    atlases: list[str],
) -> None:
    """Generate TSV files for parcellation files.

    This function will generate .tsv from:
        - the table in the .stats file
        - the secondary (commented out) information common to
          both 'left' and 'right' .stats files
    """
    info_dict = {
        InfoType.VOLUME: "GrayVol",
        InfoType.THICKNESS: "ThickAvg",
        InfoType.AREA: "SurfArea",
        InfoType.MEANCURV: "MeanCurv",
    }
    for atlas in atlases:
        hemi_to_stats_filename = _get_stats_filename_for_atlas(stats_folder, atlas)
        hemi_to_stats_df = {
            hemi: _read_stats_file(filename, ColumnType.PARCELLATION)
            for hemi, filename in hemi_to_stats_filename.items()
        }
        for info, col_name in info_dict.items():
            # Secondary information are common to left and right hemispheres
            stats_df = _get_secondary_stats(
                hemi_to_stats_filename[HemiSphere.LEFT], info
            )
            for hemi in HemiSphere:
                stats_df = pd.concat(
                    [
                        stats_df,
                        hemi_to_stats_df[hemi][["StructName", col_name]].rename(
                            columns={
                                "StructName": "label_name",
                                col_name: "label_value",
                            }
                        ),
                    ]
                )
            output_filename = (
                output_dir / f"{prefix}_parcellation-{atlas}_{info.value}.tsv"
            )
            stats_df.to_csv(output_filename, sep="\t", index=False, encoding="utf-8")


def _get_stats_filename_for_atlas(stats_folder: Path, atlas: str) -> dict:
    """Gets the stats file names for each hemisphere for requested atlas.

    Parameters
    ----------
    stats_folder : PurePath
        Path to the statistics folder.

    atlas : str
        Name of the atlas to get statistics for.

    Returns
    -------
    dict :
        Dictionary mapping hemispheres to statistics file names.
        Indexes are HemiSphere.LEFT for 'left hemisphere', and HemiSphere.RIGHT for 'right hemisphere'.
    """
    atlas_dict = {"desikan": "aparc", "destrieux": "aparc.a2009s", "ba": "BA_exvivo"}
    atlas_filename = atlas_dict.get(atlas, atlas)

    return {
        hemi: stats_folder / f"{hemi.value}.{atlas_filename}.stats"
        for hemi in HemiSphere
    }


def _read_stats_file(stats_filename: Path, column_type: ColumnType) -> pd.DataFrame:
    """Read the provided statistics file and extract the relevant data.

    Raises
    ------
    ValueError
        If `column_type` is not supported.

    FileNotFoundError
        If `stats_filename` is not a valid file name.
    """
    try:
        return pd.read_csv(
            stats_filename,
            names=_get_columns(column_type),
            comment="#",
            header=None,
            delimiter=r"\s+",
            dtype=str,
        )
    except FileNotFoundError:
        raise FileNotFoundError(f"Stats file {stats_filename} could not be found.")


def _get_columns(column_type: ColumnType) -> list[str]:
    if column_type == ColumnType.PARCELLATION:
        # Columns in ?h.BA.stats, ?h.aparc.stats or ?h.aparc.a2009s.stats file
        return [
            "StructName",
            "NumVert",
            "SurfArea",
            "GrayVol",
            "ThickAvg",
            "ThickStd",
            "MeanCurv",
            "GausCurv",
            "FoldInd",
            "CurvInd",
        ]
    if column_type == ColumnType.SEGMENTATION:
        # Columns in aseg.stats or wmparc.stats file
        return [
            "Index",
            "SegId",
            "NVoxels",
            "Volume_mm3",
            "StructName",
            "normMean",
            "normStdDev",
            "normMin",
            "normMax",
            "normRange",
        ]


def _get_secondary_stats(
    stats_filename: Path, info_type: InfoType
) -> Optional[pd.DataFrame]:
    """Read the 'secondary' statistical info from .stats file.

    Extract the information from .stats file that is commented out
    (lines starting with '# Measure' prefix) and does not appear in the
    table at the end of the document.

    Parameters
    ----------
    stats_filename : PurePath
        Path to the .stats file.

    info_type : InfoType
        Type of information to read from .stats file.

    Returns
    -------
    secondary_stats_dict : pd.DataFrame
        DataFrame containing the regions of the brain (in a column named 'label_name'),
        and the corresponding volume/thickness/area values depending on the
        input info type (in a column named 'label_value').
    """
    if info_type == InfoType.MEANCURV:
        return None
    with open(stats_filename, "r") as stats_file:
        stats = stats_file.read()
    stats_lines = [
        line for line in stats.splitlines() if _filter_stats_line(line, info_type)
    ]
    regions_stats = list(
        filter(
            None,
            (_extract_region_and_stat_value(line, info_type) for line in stats_lines),
        )
    )
    return pd.DataFrame.from_records(
        regions_stats, columns=["label_name", "label_value"]
    )


def _filter_stats_line(line: str, info: InfoType) -> bool:
    """Helper function to filter lines from stats files.

    The function returns True if the line starts with '# Measure',
    and ends with a marker specific to the type of info provided.
    More precisely, the marker is:

        - `mm^3` if info type is `volume`.
        - `mm` if  info type is `thickness`.
        - `mm^2` if info type is `area`.

    .. note::
        Info type `meancurv` is not supported.

    Parameters
    ----------
    line : str
        The line of the stat file to analyze.

    info : InfoType
        The type of information we are looking for.

    Returns
    -------
    bool :
        True if line should be kept, False otherwise.
    """
    return line.startswith("# Measure") and line.endswith(_get_end_line_marker(info))


def _get_end_line_marker(info: InfoType) -> str:
    if info == InfoType.VOLUME:
        return "mm^3"
    if info == InfoType.THICKNESS:
        return "mm"
    if info == InfoType.AREA:
        return "mm^2"
    if info == InfoType.MEANCURV:
        raise ValueError(f"InfoType cannot be {InfoType.MEANCURV.value} at this point.")


def _extract_region_and_stat_value(
    line: str, info: InfoType
) -> Optional[tuple[str, str]]:
    """Helper function to extract the region label and corresponding stat value
    from a stat line kept by `_stats_line_filter`.

    Parameters
    ----------
    line : str
        The line of the stat file to analyze.

    info : InfoType
        The type of information we are looking for.

    Returns
    -------
    region : str
        The region name for this line.

        .. note::
            This will be an empty string if the line does not contain
            specific keywords related to the type of information.

    value : str
        The corresponding statistics value.

        .. note::
            This will be an empty string if the line does not contain
            specific keywords related to the type of information.
    """
    word_list = line.replace(",", "").split()
    if any(x in word_list for x in _get_keywords(info)):
        return word_list[2], word_list[-2]
    return None


def _get_keywords(info: InfoType) -> tuple[str, str]:
    if info == InfoType.VOLUME:
        return "volume", "Volume"
    if info == InfoType.AREA:
        return "area", "Area"
    if info == InfoType.THICKNESS:
        return "thickness", "Thickness"
    if info == InfoType.MEANCURV:
        raise ValueError(f"InfoType cannot be {InfoType.MEANCURV.value} at this point.")


def _generate_tsv_for_segmentation(
    stats_folder: Path,
    output_dir: Path,
    prefix: str,
) -> None:
    """Generate TSV files for segmentation."""
    for filename, suffix in zip(
        ["aseg.stats", "wmparc.stats"],
        ["segmentationVolumes.tsv", "parcellation-wm_volume.tsv"],
    ):
        stats_filename = stats_folder / filename
        primary_stats_df = _read_stats_file(stats_filename, ColumnType.SEGMENTATION)
        primary_stats_df.rename(
            columns={"StructName": "label_name", "Volume_mm3": "label_value"},
            inplace=True,
        )
        secondary_stats_df = _get_secondary_stats(stats_filename, InfoType.VOLUME)
        full_stats_df = pd.concat(
            [
                primary_stats_df[["label_name", "label_value"]],
                secondary_stats_df,
            ]
        )
        full_stats_df.to_csv(
            output_dir / f"{prefix}_{suffix}", sep="\t", index=False, encoding="utf-8"
        )



import os
from importlib.metadata import version

from packaging.version import Version

__all__ = ["__version__", "get_version"]


__version__ = version("clinica")


def get_version() -> Version:
    return Version(__version__)


# Disable Nipype / Pydra telemetry.
os.environ["NO_ET"] = "yes"



from .. import pydra  # noqa
from . import (
    deeplearning_prepare_data,
    dwi,
    machine_learning,
    machine_learning_spatial_svm,
    pet,
    pet_surface,
    statistics_surface,
    statistics_volume,
    statistics_volume_correction,
    anatomical,
    t1_linear,
    t1_volume,
    t1_volume_create_dartel,
    t1_volume_dartel2mni,
    t1_volume_existing_template,
    t1_volume_parcellation,
    t1_volume_register_dartel,
    t1_volume_tissue_segmentation,
)



from typing import Optional

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "t1-freesurfer-longitudinal"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.overwrite_outputs
@cli_param.option.atlas_path
@option.global_option_group
@option.n_procs
@click.pass_context
def cli(
    ctx: click.Context,
    caps_directory: str,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    overwrite_outputs: bool = False,
    atlas_path: Optional[str] = None,
) -> None:
    """Longitudinal pre-processing of T1w images with FreeSurfer.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_FreeSurfer_Longitudinal/
    """
    import datetime
    from pathlib import Path

    import clinica.pipelines.anatomical.freesurfer.longitudinal.correction.cli as correction_cli
    import clinica.pipelines.anatomical.freesurfer.longitudinal.template.cli as template_cli
    from clinica.utils.longitudinal import get_participants_long_id
    from clinica.utils.participant import get_subject_session_list
    from clinica.utils.stream import cprint

    from .utils import save_part_sess_long_ids_to_tsv

    cprint(
        "The t1-freesurfer-longitudinal pipeline is divided into 2 parts:\n"
        "\tt1-freesurfer-unbiased-template pipeline: Creation of unbiased template\n"
        "\tt1-freesurfer-longitudinal-correction pipeline: Longitudinal correction."
    )
    caps_directory = Path(caps_directory)
    if not subjects_sessions_tsv:
        l_part, l_sess = get_subject_session_list(caps_directory, is_bids_dir=False)
        l_long = get_participants_long_id(l_part, l_sess)
        now = datetime.datetime.now().strftime("%H%M%S")
        subjects_sessions_tsv = f"{now}_participants.tsv"
        save_part_sess_long_ids_to_tsv(
            l_part, l_sess, l_long, Path.cwd(), subjects_sessions_tsv
        )

    cprint("Part 1/2: Running t1-freesurfer-unbiased-template pipeline.")
    ctx.invoke(
        template_cli.cli,
        caps_directory=caps_directory,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
        overwrite_outputs=overwrite_outputs,
    )

    cprint("Part 2/2 Running t1-freesurfer-longitudinal-correction pipeline.")
    ctx.invoke(
        correction_cli.cli,
        caps_directory=caps_directory,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
        overwrite_outputs=overwrite_outputs,
        atlas_path=atlas_path,
    )


if __name__ == "__main__":
    cli()



"""This module contains workflows used by the DWIPreprocessingUsingT1 pipeline."""

from typing import Optional

from nipype.pipeline.engine import Workflow

__all__ = [
    "epi_pipeline",
    "perform_ants_registration",
    "perform_dwi_epi_correction",
    "b0_flirt_pipeline",
]


def epi_pipeline(
    base_dir: str,
    delete_cache: bool = False,
    output_dir: Optional[str] = None,
    name: str = "susceptibility_distortion_correction_using_t1",
    ants_random_seed: Optional[int] = None,
    use_double_precision: bool = True,
) -> Workflow:
    """Perform EPI correction.

    This workflow allows to correct for echo-planar induced susceptibility artifacts without fieldmap
    (e.g. ADNI Database) by elastically register DWIs to their respective baseline T1-weighted
    structural scans using an inverse consistent registration algorithm with a mutual information cost
    function (SyN algorithm). This workflow allows also a coregistration of DWIs with their respective
    baseline T1-weighted structural scans in order to latter combine tracks and cortex parcellation.

    This pipeline needs:
        - FSL
        - ANTS
        - c3d

    Parameters
    ----------
    base_dir: str
        Working directory, which contains all of the intermediary data generated.

    delete_cache: bool
        If True, part of the temporary data is automatically deleted after usage.

    output_dir: str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.

    name: str, optional
        Name of the pipeline.

    ants_random_seed : int, optional
        The random seed to be used with ANTS nodes.
        If None, no random seed will be used and results will
        be stochastic.
        Default=None.

    use_double_precision : bool, optional
        This only affects tools supporting different precision settings.
        If True, computations will be made in double precision (i.e. 64 bits).
        If False, computations will be made in float precision (i.e. 32 bits).
        Default=True.

    Returns
    -------
    Workflow :
        The Nipype workflow.
        This workflow has the following inputs:
            - "t1_filename": The path to the T1w image
            - "dwi_filename": The path to the corrected DWI image
              This corresponds to the 'out_corrected' output from the eddy_fsl pipeline.
            - "b_vectors_filename": The path to the file holding the rotated B-vectors
              This corresponds to the 'out_rotated_bvecs' output from the eddy_fsl pipeline.

        And the following outputs:
            - merged_transforms
            - dwi_to_t1_co_registration_matrix
            - epi_correction_deformation_field
            - epi_correction_affine_transform
            - epi_correction_image_warped
            - rotated_b_vectors
            - epi_corrected_dwi_image


    Warnings
    --------
    This workflow rotates the b-vectors.

    Notes
    -----
    Nir et al. (2015): Connectivity network measures predict volumetric atrophy in mild cognitive impairment
    Leow et al. (2007): Statistical Properties of Jacobian Maps and the Realization of
    Unbiased Large Deformation Nonlinear Image Registration
    """
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    workflow_inputs = ["t1_filename", "dwi_filename", "b_vectors_filename"]
    workflow_outputs = ["rotated_b_vectors", "epi_corrected_dwi_image"]

    inputnode = pe.Node(
        niu.IdentityInterface(fields=workflow_inputs),
        name="inputnode",
    )
    ants_registration = perform_ants_registration(
        base_dir=base_dir,
        output_dir=output_dir,
        ants_random_seed=ants_random_seed,
    )
    epi_correction = perform_dwi_epi_correction(
        base_dir=base_dir,
        delete_cache=delete_cache,
        output_dir=output_dir,
        use_double_precision=use_double_precision,
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=workflow_outputs),
        name="outputnode",
    )

    wf = pe.Workflow(name=name)

    connections = [
        (
            inputnode,
            ants_registration,
            [(inpt, f"inputnode.{inpt}") for inpt in workflow_inputs],
        ),
        (
            inputnode,
            epi_correction,
            [
                ("dwi_filename", "inputnode.dwi_filename"),
                ("t1_filename", "inputnode.t1_filename"),
            ],
        ),
        (
            ants_registration,
            epi_correction,
            [("outputnode.merged_transforms", "inputnode.merged_transforms")],
        ),
        (
            ants_registration,
            outputnode,
            [("outputnode.rotated_b_vectors", "rotated_b_vectors")],
        ),
        (
            epi_correction,
            outputnode,
            [("outputnode.epi_corrected_dwi_image", "epi_corrected_dwi_image")],
        ),
    ]

    wf.connect(connections)

    return wf


def perform_ants_registration(
    base_dir: str,
    output_dir: Optional[str] = None,
    name: str = "perform_ants_registration",
    ants_random_seed: Optional[int] = None,
) -> Workflow:
    """Step 1 of EPI pipeline.

    Parameters
    ----------
    base_dir: str
        Working directory, which contains all the intermediary data generated.

    output_dir : str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.

    name : str, optional
        Name of the pipeline.

    ants_random_seed : int, optional
        The random seed to be used with ANTS nodes.
        If None, no random seed will be used and results will
        be stochastic.
        Default=None.

    Returns
    -------
    Workflow :
        The Nipype workflow.
        This workflow takes as inputs:
            - t1_filename : The path to the T1w input image.
            - dwi_filename : The path to the DWI image.
            - b_vectors_filename : The path to the B-vectors file.
        The workflow produces as outputs:
            - merged_transforms
            - dwi_to_t1_co_registration_matrix
            - epi_correction_deformation_field
            - epi_correction_affine_transform
            - epi_correction_image_warped
            - rotated_b_vectors
    """
    import nipype.interfaces.ants as ants
    import nipype.interfaces.c3 as c3
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.io as nio
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    from .tasks import (
        broadcast_matrix_filename_to_match_b_vector_length_task,
        change_itk_transform_type_task,
        rotate_b_vectors_task,
    )

    workflow_inputs = ["t1_filename", "dwi_filename", "b_vectors_filename"]
    workflow_outputs = [
        "merged_transforms",
        "dwi_to_t1_co_registration_matrix",
        "epi_correction_deformation_field",
        "epi_correction_affine_transform",
        "epi_correction_image_warped",
        "rotated_b_vectors",
    ]

    inputnode = pe.Node(niu.IdentityInterface(fields=workflow_inputs), name="inputnode")
    split_dwi_volumes = pe.Node(fsl.Split(dimension="t"), name="split_dwi_volumes")
    pick_reference_b0 = pe.Node(niu.Select(), name="pick_reference_b0")
    pick_reference_b0.inputs.index = [0]

    co_register_reference_b0_to_t1 = pe.Node(
        interface=fsl.FLIRT(dof=6),
        name="co_register_reference_b0_to_t1",
    )
    co_register_reference_b0_to_t1.inputs.interp = "spline"
    co_register_reference_b0_to_t1.inputs.cost = "normmi"
    co_register_reference_b0_to_t1.inputs.cost_func = "normmi"

    expend_matrix = pe.Node(
        interface=niu.Function(
            input_names=["matrix_filename", "b_vectors_filename"],
            output_names=["out_matrix_list"],
            function=broadcast_matrix_filename_to_match_b_vector_length_task,
        ),
        name="expend_matrix",
    )

    b_vectors_rotation = pe.Node(
        niu.Function(
            input_names=["matrix_filenames", "b_vectors_filename"],
            output_names=["rotated_b_vectors_filename"],
            function=rotate_b_vectors_task,
        ),
        name="b_vectors_rotation",
    )

    ants_registration = pe.Node(
        interface=ants.registration.RegistrationSynQuick(
            transform_type="br",
            dimension=3,
            precision_type="double",
            num_threads=1,
        ),
        name="antsRegistrationSyNQuick",
    )
    if ants_random_seed is not None:
        ants_registration.inputs.random_seed = ants_random_seed

    c3d_flirt2ants = pe.Node(c3.C3dAffineTool(), name="fsl_reg_2_itk")
    c3d_flirt2ants.inputs.itk_transform = True
    c3d_flirt2ants.inputs.fsl2ras = True

    change_transform_type = pe.Node(
        niu.Function(
            input_names=["input_affine_file"],
            output_names=["updated_affine_file"],
            function=change_itk_transform_type_task,
        ),
        name="change_transform_type",
    )

    merge_transforms = pe.Node(niu.Merge(3), name="merge_transforms")

    outputnode = pe.Node(
        niu.IdentityInterface(fields=workflow_outputs), name="outputnode"
    )

    if output_dir:
        write_results = pe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = output_dir
        write_results.inputs.parameterization = False

    wf = pe.Workflow(name=name)

    connections = [
        (inputnode, split_dwi_volumes, [("dwi_filename", "in_file")]),
        (split_dwi_volumes, pick_reference_b0, [("out_files", "inlist")]),
        (pick_reference_b0, co_register_reference_b0_to_t1, [("out", "in_file")]),
        (inputnode, co_register_reference_b0_to_t1, [("t1_filename", "reference")]),
        (inputnode, b_vectors_rotation, [("b_vectors_filename", "b_vectors_filename")]),
        (
            co_register_reference_b0_to_t1,
            expend_matrix,
            [("out_matrix_file", "matrix_filename")],
        ),
        (inputnode, expend_matrix, [("b_vectors_filename", "b_vectors_filename")]),
        (expend_matrix, b_vectors_rotation, [("out_matrix_list", "matrix_filenames")]),
        (inputnode, ants_registration, [("t1_filename", "fixed_image")]),
        (
            co_register_reference_b0_to_t1,
            ants_registration,
            [("out_file", "moving_image")],
        ),
        (inputnode, c3d_flirt2ants, [("t1_filename", "reference_file")]),
        (pick_reference_b0, c3d_flirt2ants, [("out", "source_file")]),
        (
            co_register_reference_b0_to_t1,
            c3d_flirt2ants,
            [("out_matrix_file", "transform_file")],
        ),
        (
            c3d_flirt2ants,
            change_transform_type,
            [("itk_transform", "input_affine_file")],
        ),
        (change_transform_type, merge_transforms, [("updated_affine_file", "in1")]),
        (ants_registration, merge_transforms, [("out_matrix", "in2")]),
        (ants_registration, merge_transforms, [("forward_warp_field", "in3")]),
        (merge_transforms, outputnode, [("out", "merged_transforms")]),
        (
            co_register_reference_b0_to_t1,
            outputnode,
            [("out_matrix_file", "dwi_to_t1_co_registration_matrix")],
        ),
        (
            ants_registration,
            outputnode,
            [
                ("forward_warp_field", "epi_correction_deformation_field"),
                ("out_matrix", "epi_correction_affine_transform"),
                ("warped_image", "epi_correction_image_warped"),
            ],
        ),
        (
            b_vectors_rotation,
            outputnode,
            [("rotated_b_vectors_filename", "rotated_b_vectors")],
        ),
    ]

    if output_dir:
        connections += [
            (
                outputnode,
                write_results,
                [(output, output) for output in workflow_outputs],
            ),
        ]

    wf.connect(connections)

    return wf


def perform_dwi_epi_correction(
    base_dir: str,
    delete_cache: bool = False,
    output_dir: Optional[str] = None,
    use_double_precision: bool = True,
    name: str = "perform_dwi_epi_correction",
) -> Workflow:
    """Step 2 of EPI pipeline.

    Parameters
    ----------
    base_dir: str
        Working directory, which contains all the intermediary data generated.

    delete_cache : bool
        If True, part of the temporary data is automatically deleted after usage.

    output_dir : str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.

    use_double_precision : bool, optional
        This only affects tools supporting different precision settings.
        If True, computations will be made in double precision (i.e. 64 bits).
        If False, computations will be made in float precision (i.e. 32 bits).
        Default=True.

    name : str, optional
        Name of the pipeline.

    Returns
    -------
    Workflow :
        The Nipype workflow.
        This workflow takes as inputs:
            - t1_filename : The path to the T1w input image.
            - dwi_filename : The path to the DWI image.
            - merged_transforms : The three transformations computed in
              the workflow `perform_ants_registration`. These are saved to
              a 5D nifti image of shape `(size_x, size_y, size_z, 1, 3)`.
        The workflow produces as outputs:
            - epi_corrected_dwi_image

    Warnings
    --------
    This workflow writes very heavy temporary files when calling AntsApplyTransforms.

    Notes
    -----
    This workflow benefits a lot from parallelization as most operations are done on
    single DWI direction.
    """
    import nipype.interfaces.ants as ants
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.io as nio
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    from clinica.utils.filemanip import delete_directories_task

    workflow_inputs = ["t1_filename", "dwi_filename", "merged_transforms"]
    workflow_outputs = ["epi_corrected_dwi_image"]

    inputnode = pe.Node(niu.IdentityInterface(fields=workflow_inputs), name="inputnode")

    split_dwi_volumes = pe.Node(fsl.Split(dimension="t"), name="split_dwi_volumes")

    apply_transform_image = pe.MapNode(
        ants.ApplyTransforms(),
        iterfield=["input_image"],
        name="warp_image",
    )
    apply_transform_image.inputs.output_image = "out_warped.nii.gz"
    apply_transform_image.inputs.print_out_composite_warp_file = False
    if not use_double_precision:
        apply_transform_image.inputs.float = True

    apply_transform_field = pe.MapNode(
        ants.ApplyTransforms(),
        iterfield=["input_image"],
        name="warp_field",
    )
    apply_transform_field.inputs.output_image = "out_warped_field.nii.gz"
    apply_transform_field.inputs.print_out_composite_warp_file = True
    if not use_double_precision:
        apply_transform_field.inputs.float = True

    jacobian = pe.MapNode(
        interface=ants.CreateJacobianDeterminantImage(),
        iterfield=["deformationField"],
        name="jacobian",
    )
    jacobian.inputs.imageDimension = 3
    jacobian.inputs.outputImage = "Jacobian_image.nii.gz"

    jacmult = pe.MapNode(
        fsl.MultiImageMaths(op_string="-mul %s"),
        iterfield=["in_file", "operand_files"],
        name="ModulateDWIs",
    )
    if not use_double_precision:
        jacmult.inputs.output_datatype = "float"

    threshold_negative = pe.MapNode(
        fsl.Threshold(thresh=0.0),
        iterfield=["in_file"],
        name="threshold_negative",
    )
    if not use_double_precision:
        threshold_negative.inputs.output_datatype = "float"

    merge_dwi_volumes = pe.Node(fsl.Merge(dimension="t"), name="merge_dwi_volumes")

    if delete_cache:
        delete_cache_node = pe.Node(
            name="delete_cache",
            interface=niu.Function(
                inputs=["directories"],
                function=delete_directories_task,
            ),
        )
        delete_cache_node.inputs.directories = [
            apply_transform_field.output_dir(),
            jacobian.output_dir(),
            jacmult.output_dir(),
            threshold_negative.output_dir(),
            apply_transform_image.output_dir(),
        ]

    outputnode = pe.Node(
        niu.IdentityInterface(fields=workflow_outputs), name="outputnode"
    )

    if output_dir:
        write_results = pe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = output_dir
        write_results.inputs.parameterization = False

    wf = pe.Workflow(name=name)

    connections = [
        (inputnode, split_dwi_volumes, [("dwi_filename", "in_file")]),
        (inputnode, apply_transform_image, [("t1_filename", "reference_image")]),
        (split_dwi_volumes, apply_transform_image, [("out_files", "input_image")]),
        (
            inputnode,
            apply_transform_image,
            [("merged_transforms", "transforms")],
        ),
        (inputnode, apply_transform_field, [("t1_filename", "reference_image")]),
        (split_dwi_volumes, apply_transform_field, [("out_files", "input_image")]),
        (inputnode, apply_transform_field, [("merged_transforms", "transforms")]),
        (apply_transform_field, jacobian, [("output_image", "deformationField")]),
        (apply_transform_image, jacmult, [("output_image", "operand_files")]),
        (jacobian, jacmult, [("jacobian_image", "in_file")]),
        (jacmult, threshold_negative, [("out_file", "in_file")]),
        (threshold_negative, merge_dwi_volumes, [("out_file", "in_files")]),
        (merge_dwi_volumes, outputnode, [("merged_file", "epi_corrected_dwi_image")]),
    ]

    if output_dir:
        connections += [
            (
                outputnode,
                write_results,
                [(output, output) for output in workflow_outputs],
            ),
        ]

    wf.connect(connections)

    return wf


def b0_flirt_pipeline(num_b0s: int, name: str = "b0_coregistration"):
    """Rigid registration of the B0 dataset onto the first volume.

    Rigid registration is achieved using FLIRT and the normalized correlation.

    Parameters
    ----------
    num_b0s : int
        Number of B0 volumes in the dataset.

    name : str, optional
        Name of the workflow.
        Default="b0_coregistration".

    Inputnode:
        in_file(str): B0 dataset.

    Outputnode
        out_b0_reg(str): The set of B0 volumes registered to the first volume.
    """
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    from nipype.interfaces import fsl

    from .tasks import merge_nifti_images_in_time_dimension_task

    inputnode = pe.Node(niu.IdentityInterface(fields=["in_file"]), name="inputnode")
    fslroi_ref = pe.Node(fsl.ExtractROI(args="0 1"), name="b0_reference")
    tsize = num_b0s - 1
    fslroi_moving = pe.Node(fsl.ExtractROI(args="1 " + str(tsize)), name="b0_moving")
    split_moving = pe.Node(fsl.Split(dimension="t"), name="split_b0_moving")

    bet_ref = pe.Node(fsl.BET(frac=0.3, mask=True, robust=True), name="bet_ref")

    dilate = pe.Node(
        fsl.maths.MathsCommand(nan2zeros=True, args="-kernel sphere 5 -dilM"),
        name="mask_dilate",
    )

    flirt = pe.MapNode(
        fsl.FLIRT(
            interp="spline",
            dof=6,
            bins=50,
            save_log=True,
            cost="corratio",
            cost_func="corratio",
            padding_size=10,
            searchr_x=[-4, 4],
            searchr_y=[-4, 4],
            searchr_z=[-4, 4],
            fine_search=1,
            coarse_search=10,
        ),
        name="b0_co_registration",
        iterfield=["in_file"],
    )

    merge = pe.Node(fsl.Merge(dimension="t"), name="merge_registered_b0s")
    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0), iterfield=["in_file"], name="remove_negative"
    )
    insert_ref = pe.Node(
        niu.Function(
            input_names=["image1", "image2"],
            output_names=["out_file"],
            function=merge_nifti_images_in_time_dimension_task,
        ),
        name="concat_ref_moving",
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["out_file", "out_xfms"]), name="outputnode"
    )

    wf = pe.Workflow(name=name)
    connections = [
        (inputnode, fslroi_ref, [("in_file", "in_file")]),
        (inputnode, fslroi_moving, [("in_file", "in_file")]),
        (fslroi_moving, split_moving, [("roi_file", "in_file")]),
        (fslroi_ref, bet_ref, [("roi_file", "in_file")]),
        (bet_ref, dilate, [("mask_file", "in_file")]),
        (dilate, flirt, [("out_file", "ref_weight"), ("out_file", "in_weight")]),
        (fslroi_ref, flirt, [("roi_file", "reference")]),
        (split_moving, flirt, [("out_files", "in_file")]),
        (flirt, thres, [("out_file", "in_file")]),
        (thres, merge, [("out_file", "in_files")]),
        (merge, insert_ref, [("merged_file", "image2")]),
        (fslroi_ref, insert_ref, [("roi_file", "image1")]),
        (insert_ref, outputnode, [("out_file", "out_file")]),
        (flirt, outputnode, [("out_matrix_file", "out_xfms")]),
    ]
    wf.connect(connections)

    return wf



"""Module for converting FLAIR of ADNI."""

from pathlib import Path
from typing import Any, Iterable, Optional

import pandas as pd

__all__ = ["convert_flair"]


def convert_flair(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert FLAIR images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    destination_dir : Path
        The path to the destination BIDS directory.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        f"Calculating paths of {ADNIModalityConverter.FLAIR} images. Output will be stored in {conversion_dir}.",
        lvl="info",
    )
    images = _compute_flair_paths(source_dir, csv_dir, subjects, conversion_dir)
    cprint(
        f"Paths of {ADNIModalityConverter.FLAIR} images found. Exporting images into BIDS ...",
        lvl="info",
    )
    # flair_paths_to_bids(images, dest_dir)
    paths_to_bids(
        images,
        destination_dir,
        ADNIModalityConverter.FLAIR,
        mod_to_update=mod_to_update,
        n_procs=n_procs,
    )
    cprint(msg="FLAIR conversion done.", lvl="debug")


def _compute_flair_paths(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
) -> pd.DataFrame:
    """Compute the paths to the FLAIR images and store them in a TSV file.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    subjects : list of str
        The subjects list.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    Returns
    -------
    images : pd.DataFrame
        A dataframe with all the paths to the FLAIR images that will be converted into BIDS.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv

    from ._image_path_utils import find_image_path
    from ._visits_utils import visits_to_timepoints

    flair_df = _initialize_flair_df()
    flair_dfs_list = []
    # Loading needed .csv files
    adni_merge = load_clinical_csv(csv_dir, "ADNIMERGE")
    mayo_mri_qc = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15")
    mri_list = load_clinical_csv(csv_dir, "MRILIST")

    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "AFL"]

    # Selecting FLAIR DTI images that are not MPR
    mri_list = mri_list[mri_list.SEQUENCE.str.contains("flair", case=False, na=False)]
    mri_list = mri_list[
        mri_list.SEQUENCE.map(lambda x: not any(subs in x for subs in ("_MPR_",)))
    ]
    for subject in subjects:
        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subject]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mri_list_subj = mri_list[mri_list.SUBJECT == subject]
        mri_list_subj = mri_list_subj.sort_values("SCANDATE")

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subject[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subject, mri_list_subj, adnimerge_subj, "FLAIR")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]
            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            flair = _select_flair_image(
                subject, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj
            )
            if flair is not None:
                row_to_append = pd.DataFrame(flair, index=["i"])
                flair_dfs_list.append(row_to_append)
    if flair_dfs_list:
        flair_df = pd.concat(flair_dfs_list, ignore_index=True)

    # Removing known exceptions from images to convert
    if not flair_df.empty:
        error_ind = flair_df.index[
            flair_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in _get_known_conversion_errors()),
                axis=1,
            )
        ]
        flair_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(flair_df, source_dir, "FLAIR")
    images.to_csv(conversion_dir / "flair_paths.tsv", sep="\t", index=False)

    return images


def _initialize_flair_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            "Subject_ID",
            "VISCODE",
            "Visit",
            "Sequence",
            "Scan_Date",
            "Study_ID",
            "Series_ID",
            "Image_ID",
            "Field_Strength",
            "Scanner",
        ]
    )


def _get_known_conversion_errors() -> Iterable[tuple[str, str]]:
    return [
        ("141_S_0767", "m84"),
        ("067_S_5205", "bl"),
        ("127_S_4928", "m24"),
        ("024_S_4674", "m06"),
        ("123_S_2363", "m24"),
        ("053_S_4578", "m48"),
        ("128_S_4586", "m48"),
        ("053_S_4813", "m48"),
        ("053_S_5272", "m24"),
        ("013_S_1186", "m48"),
        ("031_S_2022", "bl"),
        ("031_S_2022", "m06"),
        ("031_S_2233", "bl"),
        ("031_S_2233", "m03"),
        ("029_S_2395", "m72"),
        ("130_S_6043", "bl"),
        ("031_S_2018", "bl"),
        ("027_S_5170", "m72"),
        ("135_S_6284", "m12"),
        ("068_S_0127", "m180"),
        ("068_S_2187", "m120"),
        # Several output images
        ("114_S_6039", "bl"),
    ]


def _select_flair_image(
    subject_id: str,
    timepoint,
    visit_str: str,
    visit_mri_list: list,
    mri_qc_subj: pd.DataFrame,
) -> Optional[dict[str, Any]]:
    """Select a FLAIR image based on QC information.

    One image among those in the input list is chosen according to QC
    and then corresponding metadata is extracted to a dictionary

    Parameters
    ----------
    subject_id : str
        The subject identifier.

    timepoint : ??
        The visit code.

    visit_str : str
        The visit name.

    visit_mri_list : list
        A list of images metadata.

    mri_qc_subj : pd.DataFrame
        A Dataframe containing list of QC of scans for the subject.

    Returns
    -------
    dict or None :
        Contains image metadata.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars

    from ._qc_utils import select_image_qc

    selected_image_id = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if selected_image_id is None:
        return None
    selected_scan = visit_mri_list[visit_mri_list.IMAGEUID == selected_image_id].iloc[0]

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": replace_sequence_chars(selected_scan.SEQUENCE),
        "Scan_Date": selected_scan["SCANDATE"],
        "Study_ID": str(int(selected_scan.STUDYID)),
        "Series_ID": str(int(selected_scan.SERIESID)),
        "Image_ID": str(int(selected_scan.IMAGEUID)),
        "Field_Strength": selected_scan.MAGSTRENGTH,
    }



import os

from nibabel.nifti1 import Nifti1Image
from pydra.mark import annotate, task


@task
@annotate({"return": {"transforms_list": list}})
def concatenate_transforms_task(
    pet_to_t1w_tranform: str, t1w_to_mni_tranform: str
) -> list:
    """Pydra task for concatenating two input transformation files into a list.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_linear.pet_linear_utils.concatenate_transforms`.

    """
    from clinica.pipelines.pet.linear.utils import concatenate_transforms

    return concatenate_transforms(t1w_to_mni_tranform, pet_to_t1w_tranform)


@task
@annotate({"return": {"output_img": str}})
def suvr_normalization_task(
    input_img: os.PathLike,
    norm_img: os.PathLike,
    ref_mask: os.PathLike,
) -> str:
    """Pydra task for normalizing the input image according to the reference region.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_linear.pet_linear_utils.suvr_normalization`.

    """
    from pathlib import Path

    from clinica.pipelines.pet.linear.utils import perform_suvr_normalization

    return str(
        perform_suvr_normalization(Path(input_img), Path(norm_img), Path(ref_mask))
    )


@task
@annotate({"return": {"output_img": os.PathLike}})
def crop_nifti_task(input_img: os.PathLike, output_dir: os.PathLike) -> os.PathLike:
    """Pydra task for cropping the input image based on the reference.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_linear.pet_linear_utils.crop_nifti`.

    """
    from pathlib import Path

    from clinica.utils.image import crop_nifti

    return crop_nifti(Path(input_img), Path(output_dir))



def apply_binary_mask_task(image: str, binary_mask: str) -> str:
    from pathlib import Path

    from clinica.pipelines.pet.volume.utils import apply_binary_mask

    return str(apply_binary_mask(Path(image), Path(binary_mask)))


def compute_atlas_statistics_task(image: str, atlas_names: list) -> list:
    from pathlib import Path

    from clinica.pipelines.pet.volume.utils import compute_atlas_statistics

    return [str(p) for p in compute_atlas_statistics(Path(image), atlas_names)]


def create_binary_mask_task(
    tissues: list,
    threshold: float = 0.3,
) -> str:
    from pathlib import Path

    from clinica.pipelines.pet.volume.utils import create_binary_mask

    return str(create_binary_mask([Path(tissue) for tissue in tissues], threshold))


def create_pvc_mask_task(tissues: list) -> str:
    from pathlib import Path

    from clinica.pipelines.pet.volume.utils import create_pvc_mask

    return str(create_pvc_mask([Path(tissue) for tissue in tissues]))


def normalize_to_reference_task(pet_image: str, region_mask: str) -> str:
    from pathlib import Path

    from clinica.pipelines.pet.volume.utils import normalize_to_reference

    return str(normalize_to_reference(Path(pet_image), Path(region_mask)))



from typing import List


def peak_correction(t_map: str, t_threshold: float, output_name: str = None) -> str:
    """Threshold the t_map with t_threshold.

    Pixel intensities that are less than t_threshold are set to 0, other values
    are left unchanged.

    Parameters
    ----------
    t_map: str
        Path to t-statistics nifti map
    t_threshold: float
        Threshold on t value
    output_name: str, optional
        Optional output name

    Returns
    -------
    str:
        Path to the generated file.
    """
    from os.path import abspath, basename, join

    import nibabel as nib

    original_nifti = nib.load(t_map)
    data = original_nifti.get_fdata(dtype="float32")
    data[data < t_threshold] = 0
    new_data = nib.Nifti1Image(
        data, affine=original_nifti.affine, header=original_nifti.header
    )
    if output_name:
        filename = output_name
    else:
        filename = join("./peak_corrected_" + str(t_threshold) + basename(t_map))
    nib.save(new_data, filename)
    return abspath(filename)


def cluster_correction(
    t_map: str, t_thresh: float, c_thresh: int, output_name: str = None
) -> str:
    """Performs cluster correction.

    First t_map is thresholded with t_thresh (like in peak_correction()). Then, clusters
    that have a size less than c_thresh are removed

    Parameters
    ----------
    t_map: str
        Path to t-statistics nifti map
    t_thresh: float
        Threshold on t value
    c_thresh: int
        Minimal size of clusters after thresholding
    output_name: str, optional
        Optional output name

    Returns
    -------
    str:
        Path to the generated file.
    """
    from os.path import abspath, basename, join

    import nibabel as nib
    import numpy as np
    from scipy.ndimage.measurements import label

    original_nifti = nib.load(t_map)
    data = original_nifti.get_fdata(dtype="float32")
    data[data < t_thresh] = 0
    labeled_mask, num_features = label(data)
    for i in range(1, num_features + 1):
        if np.sum(labeled_mask == i) < c_thresh:
            print(
                "Label number "
                + str(i)
                + " cluster size is: "
                + str(np.sum(labeled_mask == i))
                + " so it is removed"
            )
            data[labeled_mask == i] = 0
    new_data = nib.Nifti1Image(
        data, affine=original_nifti.affine, header=original_nifti.header
    )
    if output_name:
        filename = output_name
    else:
        filename = join(
            "./cluster_corrected_t-"
            + str(t_thresh)
            + "_c-"
            + str(c_thresh)
            + basename(t_map)
        )
    nib.save(new_data, filename)
    return abspath(filename)


def produce_figures(
    nii_file: str,
    template: str,
    type_of_correction: str,
    t_thresh: str,
    c_thresh: int,
    n_cuts: int,
) -> list:
    """Produce the output figures.

    Parameters
    ----------
    nii_file: str
        Path to the nifti file (generated at previous steps)
    template: str
        Path to template used for the stat map plot
    type_of_correction: str
        Can be either FWE or FDR (used only in potential figure titles)
    t_thresh: str
        T value threshold used (used only in potential figure titles)
    c_thresh: int
        Cluster minimal size used (used only in potential figure titles)
    n_cuts: int
        Number of cuts in fig

    Returns
    -------
    list of string:
        List of path to image files: glass brain, statmap along x, statmap along y, statmap along z
    """
    from os.path import abspath

    import numpy as np
    from nilearn import plotting

    assert type_of_correction in ["FWE", "FDR"], "Type of correction must be FWE or FDR"
    if not np.isnan(c_thresh):
        correction = "Cluster"
    else:
        correction = "Peak"

    my_title = (
        correction
        + " correction "
        + type_of_correction
        + " Threshold = "
        + str(t_thresh)
    )
    if not np.isnan(c_thresh):
        my_title = (my_title + " - min cluster size = " + str(c_thresh),)

    plotting.plot_glass_brain(nii_file, output_file="./glass_brain.png")

    plotting.plot_stat_map(
        nii_file,
        display_mode="x",
        cut_coords=np.linspace(-70, 67, n_cuts),
        bg_img=template,
        colorbar=False,
        draw_cross=True,
        output_file="./statmap_x.png",
    )

    plotting.plot_stat_map(
        nii_file,
        display_mode="y",
        cut_coords=np.linspace(-104, 69, n_cuts),
        bg_img=template,
        colorbar=False,
        draw_cross=True,
        output_file="./statmap_y.png",
    )

    plotting.plot_stat_map(
        nii_file,
        display_mode="z",
        cut_coords=np.linspace(-45, 78, n_cuts),
        bg_img=template,
        colorbar=False,
        draw_cross=True,
        output_file="./statmap_z.png",
    )

    return [
        abspath("./glass_brain.png"),
        abspath("./statmap_x.png"),
        abspath("./statmap_y.png"),
        abspath("./statmap_z.png"),
    ]


def generate_output(t_map: str, figs: list, correction_name: str) -> None:
    """Extract the output generated and copy it to the output folder

    Parameters
    ----------
    t_map: str
        Path to t-map on which whole pipeline was based
    figs: list of str
        Paths to figs to save
    correction_name: str
        Name of the correction (ex: cluster_correction_FWE)
    """
    from os import makedirs
    from os.path import basename, dirname, join, splitext
    from shutil import copyfile

    # Will extract group-GroupTest_AD-lt-CN_measure-fdg_fwhm-8_TStatistics from TStatistics file
    t_map_basename = splitext(basename(t_map))[0]

    out_folder = join(
        dirname(t_map), t_map_basename.replace("TStatistics", correction_name)
    )
    makedirs(out_folder)
    suffixes = (
        "GlassBrain.png",
        "axis-x_TStatistics.png",
        "axis-y_TStatistics.png",
        "axis-z_TStatistics.png",
    )
    for fig, suffix in zip(figs, suffixes):
        copyfile(
            fig,
            join(
                out_folder,
                t_map_basename.replace(
                    "TStatistics", f"desc-{correction_name}_{suffix}"
                ),
            ),
        )



from typing import Optional, Union

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline
from clinica.utils.atlas import T1AndPetVolumeAtlasName
from clinica.utils.pet import SUVRReferenceRegion, Tracer

pipeline_name = "machinelearning-classification"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@click.argument(
    "orig_input_data",
    type=click.Choice(["VoxelBased", "RegionBased"]),
)
@click.argument(
    "image_type",
    type=click.Choice(["T1w", "PET"]),
)
@click.argument(
    "algorithm",
    type=click.Choice(["DualSVM", "LogisticRegression", "RandomForest"]),
)
@click.argument(
    "validation",
    type=click.Choice(["RepeatedHoldOut", "RepeatedKFoldCV"]),
)
@click.argument(
    "subjects_visits_tsv",
    type=click.Path(exists=True, resolve_path=True),
)
@click.argument(
    "diagnoses_tsv",
    type=click.Path(exists=True, resolve_path=True),
)
@click.argument(
    "output_directory", type=click.Path(exists=False, writable=True, resolve_path=True)
)
@cli_param.option_group.pipeline_specific_options
@cli_param.option.acq_label
@cli_param.option.suvr_reference_region
@cli_param.option_group.option(
    "-atlas",
    "--atlas",
    type=click.Choice(T1AndPetVolumeAtlasName),
    help="One of the atlases generated by t1-volume or pet-volume pipeline.",
)
@option.global_option_group
@option.n_procs
def cli(
    caps_directory: str,
    group_label: str,
    orig_input_data: str,
    image_type: str,
    algorithm: str,
    validation: str,
    subjects_visits_tsv: str,
    diagnoses_tsv: str,
    output_directory: str,
    acq_label: Optional[Union[str, Tracer]] = None,
    suvr_reference_region: Optional[Union[str, SUVRReferenceRegion]] = None,
    atlas: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Classification based on machine learning using scikit-learn.

    Parameters
    ----------
    caps_directory : str

    group_label : str
        String defining the group label for the current analysis, which helps you keep track of different analyses.

    orig_input_data : str
        Defines the type of features for classification.
        It can be 'RegionBased' or 'VoxelBased'.

    image_type : str
        Defines the studied modality ('T1w' or 'PET')

    algorithm : str
        Defines the algorithm. It can be 'DualSVM', 'LogisticRegression' or 'RandomForest'.

    validation : str
        Defines the validation method. It can be 'RepeatedHoldOut' or 'RepeatedKFoldCV'.

    subjects_visits_tsv : str
        TSV file containing the participant_id and the session_id columns.

    diagnoses_tsv : str
        TSV file where the diagnosis for each participant (identified by a participant ID) is reported (e.g. AD, CN).
        It allows the algorithm to perform the dual classification (between the two labels reported).

    output_directory : str
        The output folder path.

    acq_label : str, optional

    suvr_reference_region : str, optional

    atlas : str, optional

    n_procs : int, optional
        The number of processes to be used by the pipeline.

    Notes
    -----
    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/MachineLearning_Classification/
    """
    from clinica.utils.exceptions import ClinicaException

    from .ml_workflows import (
        RegionBasedRepHoldOutDualSVM,
        RegionBasedRepHoldOutLogisticRegression,
        RegionBasedRepHoldOutRandomForest,
        RegionBasedRepKFoldDualSVM,
        VoxelBasedRepHoldOutDualSVM,
        VoxelBasedRepKFoldDualSVM,
    )

    if image_type == "PET":
        if not acq_label:
            raise ClinicaException(
                "You selected PET inputs without setting --acq_label flag. "
                "Clinica will now exit."
            )
        if not suvr_reference_region:
            raise ClinicaException(
                "You selected PET inputs without setting --suvr_reference_region flag. "
                "Clinica will now exit."
            )

    if orig_input_data == "RegionBased" and not atlas:
        raise ClinicaException(
            "You selected region-based inputs without setting --atlas flag. "
            "Clinica will now exit."
        )

    if algorithm in ("LogisticRegression", "RandomForest"):
        if orig_input_data != "RegionBased" or validation != "RepeatedHoldOut":
            raise ClinicaException(
                "LogisticRegression or RandomForest algorithm can only work "
                "on region-based featured or RepeatedHoldOut algorithm. "
                "Clinica will now exit."
            )

    if (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "DualSVM")
    ):
        pipeline = RegionBasedRepHoldOutDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedKFoldCV")
        and (algorithm == "DualSVM")
    ):
        pipeline = RegionBasedRepKFoldDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "LogisticRegression")
    ):
        pipeline = RegionBasedRepHoldOutLogisticRegression(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "RandomForest")
    ):
        pipeline = RegionBasedRepHoldOutRandomForest(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "VoxelBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "DualSVM")
    ):
        pipeline = VoxelBasedRepHoldOutDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "VoxelBased")
        and (validation == "RepeatedKFoldCV")
        and (algorithm == "DualSVM")
    ):
        pipeline = VoxelBasedRepKFoldDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            n_threads=n_procs,
        )
    else:
        raise ClinicaException(
            "Unknown combination of machine learning classification."
        )

    pipeline.run()


if __name__ == "__main__":
    cli()



from dataclasses import dataclass

import numpy as np
import pandas as pd
from brainstat.stats.SLM import SLM

from ._base import Results

__all__ = ["StatisticsResults"]


@dataclass
class PValueResults(Results):
    """This class implements a container for raw (uncorrected)
    P-value results obtained with a GLM model.

    Attributes
    ----------
    pvalues : np.ndarray
        Array of uncorrected P-values.

    mask : np.ndarray
        The binary mask.

    threshold : float
        The threshold used.
    """

    pvalues: np.ndarray
    mask: np.ndarray
    threshold: float

    @property
    def thresh(self):
        """For compatibility with previous Matlab implementation."""
        return self.threshold

    @property
    def P(self):
        """For compatibility with previous Matlab implementation."""
        return self.pvalues

    @classmethod
    def from_t_statistics(
        cls,
        tstats: np.ndarray,
        df: pd.DataFrame,
        mask: np.ndarray,
        threshold: float,
    ):
        """Instantiate the class from an array of T-statistics.

        Parameters
        ----------
        tstats : np.ndarray
            Array of T-statistics.

        df : pd.DataFrame
            The subjects DataFrame.

        mask : np.ndarray
            The binary mask.

        threshold : float
            The threshold to be used.
        """
        from scipy.stats import t

        return cls(1 - t.cdf(tstats, df), mask, threshold)


@dataclass
class CorrectedPValueResults(PValueResults):
    """This class implements a container for corrected P-value
    results obtained with a GLM model.

    Attributes
    ----------
    cluster_pvalues : np.ndarray
        The cluster P-values.
    """

    cluster_pvalues: np.ndarray

    @property
    def C(self):
        """For compatibility with previous Matlab implementation."""
        return self.cluster_pvalues


@dataclass
class StatisticsResults(Results):
    """This class implements a container for results obtained with
    the GLM model classes. It holds information relative to a GLM
    run with one specific contrast.

    Attributes
    ----------
    coefficients : np.ndarray
        The beta coefficients of the fitted GLM model.

    tstats : np.ndarray
        The corresponding T-statistics.

    uncorrected_p_value : PValueResults
        The corresponding uncorrected p values, stored in a `PValueResults` instance.

    fdr : np.ndarray
        The corresponding False Discovery Rate.

    corrected_p_value : CorrectedPValueResults
        The corresponding corrected p values, stored in a `CorrectedPValueResults` instance.
    """

    coefficients: np.ndarray
    tstats: np.ndarray
    uncorrected_p_values: PValueResults
    fdr: np.ndarray
    corrected_p_values: CorrectedPValueResults

    @property
    def TStatistics(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.tstats

    @property
    def uncorrectedPValue(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.uncorrected_p_values

    @property
    def correctedPValue(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.corrected_p_values

    @property
    def FDR(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.fdr

    @classmethod
    def from_slm_model(
        cls,
        model: SLM,
        mask: np.ndarray,
        threshold_uncorrected_p_value: float,
        threshold_corrected_p_value: float,
    ):
        """Instantiate from a SLM model.

        Parameters
        ----------
        model : brainstat.stats.SLM
            SLM model instance to use.

        mask : np.ndarray
            The binary mask to use.

        threshold_uncorrected_p_value : float
            The threshold to use with uncorrected P-values.

        threshold_corrected_p_value : float
            The threshold to use with corrected P-values.
        """
        idx = np.argwhere(np.isnan(model.t))
        corrected_pvals = model.P["pval"]["P"]
        corrected_pvals[idx] = 1.0
        tstats = np.nan_to_num(model.t)
        uncorrected_p_values = PValueResults.from_t_statistics(
            tstats,
            model.df,
            mask,
            threshold_uncorrected_p_value,
        )
        corrected_p_values = CorrectedPValueResults(
            corrected_pvals,
            mask,
            threshold_corrected_p_value,
            model.P["pval"]["C"],
        )
        return cls(
            np.nan_to_num(model.coef),
            tstats,
            uncorrected_p_values,
            model.Q,
            corrected_p_values,
        )



from os import PathLike
from pathlib import Path
from typing import Optional

from pydra.mark import annotate, task


@task
@annotate({"return": {"output_file": Path}})
def write_bids_file(
    input_file: PathLike,
    dataset_path: PathLike,
    participant_id: str,
    session_id: Optional[str],
    datatype: str,
    suffix: str,
    entities: dict,
) -> Path:
    """Task to write a file to a BIDS dataset.

    Examples
    --------
    >>> from shutil import rmtree
    >>> from tempfile import mkdtemp, mkstemp
    >>> test_input_file = Path(mkstemp(suffix=".nii.gz")[1])
    >>> test_output_dir = Path(mkdtemp())
    >>> task = write_bids_file(
    ...     input_file=test_input_file,
    ...     dataset_path=test_output_dir,
    ...     participant_id="P01",
    ...     session_id="M00",
    ...     datatype="anat",
    ...     suffix="T1w",
    ...     entities={"desc": "copy"},
    ... )
    >>> result = task()
    >>> str(result.output.output_file.relative_to(test_output_dir))
    'sub-P01/ses-M00/anat/sub-P01_ses-M00_desc-copy_T1w.nii.gz'
    >>> str(next(test_output_dir.rglob("*.nii.gz")).relative_to(test_output_dir))
    'sub-P01/ses-M00/anat/sub-P01_ses-M00_desc-copy_T1w.nii.gz'
    >>> test_input_file.unlink()
    >>> rmtree(test_output_dir)
    """
    input_file = Path(input_file)
    input_ext = input_file.name.split(sep=".", maxsplit=1)[-1]

    output_dir = (
        Path(dataset_path)
        / f"sub-{participant_id}"
        / (f"ses-{session_id}" if session_id else "")
        / datatype
    )

    bids_stem = "_".join(
        [f"sub-{participant_id}"]
        + ([f"ses-{session_id}"] if session_id else [])
        + ([f"{ent}-{val}" for ent, val in entities.items()] if entities else [])
        + [suffix]
    )

    output_file = output_dir / ".".join([bids_stem, input_ext])

    # Ensure target directory exists.
    output_dir.mkdir(parents=True, exist_ok=True)

    # Write source file to target directory.
    output_file.write_bytes(input_file.read_bytes())

    return output_file


@task
@annotate({"return": {"output_file": Path}})
def write_caps_file(
    input_file: PathLike,
    dataset_path: PathLike,
    participant_id: str,
    session_id: Optional[str],
    datatype: str,
    suffix: str,
    entities: dict,
) -> Path:
    """Task to write a file a CAPS dataset.

    Examples
    --------
    >>> from shutil import rmtree
    >>> from tempfile import mkdtemp, mkstemp
    >>> test_input_file = Path(mkstemp(suffix=".mat")[1])
    >>> test_output_dir = Path(mkdtemp())
    >>> task = write_caps_file(
    ...     input_file=test_input_file,
    ...     dataset_path=test_output_dir,
    ...     participant_id="P01",
    ...     session_id="M00",
    ...     datatype="anat",
    ...     suffix="space-{space}_res-{res}_affine",
    ...     entities={"space": "MNI152", "res": "1x1x1"},
    ... )
    >>> result = task()
    >>> str(result.output.output_file.relative_to(test_output_dir))
    'sub-P01/ses-M00/anat/sub-P01_ses-M00_space-MNI152_res-1x1x1_affine.mat'
    >>> str(next(test_output_dir.rglob("*.mat")).relative_to(test_output_dir))
    'sub-P01/ses-M00/anat/sub-P01_ses-M00_space-MNI152_res-1x1x1_affine.mat'
    >>> test_input_file.unlink()
    >>> rmtree(test_output_dir)
    """
    input_file = Path(input_file)
    input_ext = input_file.name.split(sep=".", maxsplit=1)[-1]

    output_dir = (
        Path(dataset_path)
        / f"sub-{participant_id}"
        / (f"ses-{session_id}" if session_id else "")
        / datatype.format(**entities)
    )

    caps_stem = "_".join(
        [f"sub-{participant_id}"]
        + ([f"ses-{session_id}"] if session_id else [])
        + [suffix.format(**entities)]
    )

    output_file = output_dir / ".".join([caps_stem, input_ext])

    # Ensure target directory exists.
    output_dir.mkdir(parents=True, exist_ok=True)

    # Write source file to target directory.
    output_file.write_bytes(input_file.read_bytes())

    return output_file



"""Module for converting AV45 and Florbetaben PET of ADNI."""

from pathlib import Path
from typing import Iterable

import pandas as pd

__all__ = ["convert_av45_fbb_pet"]


def convert_av45_fbb_pet(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert AV-45 and Florbetaben PET images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    destination_dir : Path
        The path to the destination BIDS directory.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, default=1
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        f"Calculating paths of AV45 and Florbetaben PET images. Output will be stored in {conversion_dir}."
    )
    images = _compute_av45_fbb_pet_paths(source_dir, csv_dir, subjects, conversion_dir)
    cprint(
        "Paths of AV45 and Florbetaben PET images found. Exporting images into BIDS ..."
    )
    paths_to_bids(
        images,
        destination_dir,
        ADNIModalityConverter.PET_AV45,
        mod_to_update=mod_to_update,
        n_procs=n_procs,
    )
    cprint(msg="AV45 and Florbetaben PET conversion done.", lvl="debug")


def _compute_av45_fbb_pet_paths(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
) -> pd.DataFrame:
    """Compute the paths to the AV45 and Florbetaben PET images and store them in a TSV file.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    subjects : list of str
        The subjects list.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    Returns
    -------
    images : pd.DataFrame
        Dataframe with all the paths to the PET images that will be converted into BIDS.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv

    from ._image_path_utils import find_image_path
    from ._pet_utils import get_images_pet

    pet_amyloid_col = [
        "Phase",
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Original",
        "Tracer",
    ]
    pet_amyloid_df = pd.DataFrame(columns=pet_amyloid_col)
    pet_amyloid_dfs_list = []

    # Loading needed .csv files
    av45qc = load_clinical_csv(csv_dir, "AV45QC")
    amyqc = load_clinical_csv(csv_dir, "AMYQC")
    pet_meta_list = load_clinical_csv(csv_dir, "PET_META_LIST")

    for subject in subjects:
        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list["Subject"] == subject]

        if subject_pet_meta.empty:
            continue
        # QC for AV45 PET images for ADNI 1, GO and 2
        av45_qc_subj = av45qc[(av45qc.PASS == 1) & (av45qc.RID == int(subject[-4:]))]

        # QC for Amyloid PET images for ADNI 3
        amy_qc_subj = amyqc[(amyqc.SCANQLTY == 1) & (amyqc.RID == int(subject[-4:]))]
        amy_qc_subj.insert(0, "EXAMDATE", amy_qc_subj.SCANDATE.to_list())

        # Concatenating visits in both QC files
        amyloid_qc_subj = pd.concat(
            [av45_qc_subj, amy_qc_subj], axis=0, ignore_index=True, sort=False
        )
        sequences_preprocessing_step = [
            "AV45 Co-registered, Averaged",
            "FBB Co-registered, Averaged",
        ]
        subj_dfs_list = get_images_pet(
            subject,
            amyloid_qc_subj,
            subject_pet_meta,
            pet_amyloid_col,
            "Amyloid-PET",
            sequences_preprocessing_step,
        )
        if subj_dfs_list:
            pet_amyloid_dfs_list += subj_dfs_list

    if pet_amyloid_dfs_list:
        pet_amyloid_df = pd.concat(pet_amyloid_dfs_list, ignore_index=True)

    # Removing known exceptions from images to convert
    if not pet_amyloid_df.empty:
        error_ind = pet_amyloid_df.index[
            pet_amyloid_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in _get_known_conversion_errors()),
                axis=1,
            )
        ]
        pet_amyloid_df.drop(error_ind, inplace=True)

    images = find_image_path(pet_amyloid_df, source_dir, "Amyloid")
    images.to_csv(conversion_dir / "amyloid_pet_paths.tsv", sep="\t", index=False)

    return images


def _get_known_conversion_errors() -> Iterable[tuple[str, str]]:
    return [
        ("128_S_2220", "m48"),
        # Several output images
        ("098_S_4275", "m84"),
    ]



# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
from typing import List

from nipype import config

from clinica.pipelines.pet.engine import PETPipeline

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class PETLinear(PETPipeline):
    """PET Linear - Affine registration of PET images to standard space.
    This preprocessing pipeline uses T1w MRI transformation into standard
    space computed in clinica t1-linear pipeline.
    It includes globally four steps:
    1) PET image registration into associated T1w MRI image space
        with RegistrationSynQuick from ANTs.
    2) PET image registration to MNI152NLin2009cSym template with
       RegistrationSynQuick from ANTs.
    3) SUVR voxel intensity normalisation using Pons Cerebellum masks.
    4) Crop the background (in order to save computational power).

    Returns:
        A clinica pipeline object containing the pet_linear pipeline.
    """

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["pet", "t1w", "t1w_to_mni"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return [
            "registered_pet",
            "transform_mat",
            "registered_pet_in_t1w",
        ]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.pipelines.pet.utils import get_suvr_mask
        from clinica.utils.exceptions import (
            ClinicaBIDSError,
            ClinicaCAPSError,
            ClinicaException,
        )
        from clinica.utils.image import get_mni_template
        from clinica.utils.input_files import T1W_NII, T1W_TO_MNI_TRANSFORM
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        self.ref_template = get_mni_template("t1")
        self.ref_mask = get_suvr_mask(self.parameters["suvr_reference_region"])

        # Inputs from BIDS directory
        try:
            pet_files, _ = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.bids_directory,
                self._get_pet_scans_query(),
            )
        except ClinicaException as e:
            err = (
                "Clinica faced error(s) while trying to read pet files in your BIDS directory.\n"
                + str(e)
            )
            raise ClinicaBIDSError(err)

        # T1w file:
        try:
            t1w_files, _ = clinica_file_reader(
                self.subjects, self.sessions, self.bids_directory, T1W_NII
            )
        except ClinicaException as e:
            err = (
                "Clinica faced error(s) while trying to read t1w files in your BIDS directory.\n"
                + str(e)
            )
            raise ClinicaBIDSError(err)

        # Inputs from t1-linear pipeline
        # Transformation files from T1w files to MNI:
        try:
            t1w_to_mni_transformation_files, _ = clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, T1W_TO_MNI_TRANSFORM
            )
        except ClinicaException as e:
            err = (
                "Clinica faced error(s) while trying to read transformation files in your CAPS directory.\n"
                + str(e)
            )
            raise ClinicaCAPSError(err)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint("The pipeline will last approximately 3 minutes per image.")

        read_input_node = npe.Node(
            name="LoadingCLIArguments",
            iterables=[
                ("t1w", t1w_files),
                ("pet", pet_files),
                ("t1w_to_mni", t1w_to_mni_transformation_files),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_input_node, self.input_node, [("t1w", "t1w")]),
                (read_input_node, self.input_node, [("pet", "pet")]),
                (read_input_node, self.input_node, [("t1w_to_mni", "t1w_to_mni")]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.io import DataSink

        from clinica.utils.nipype import container_from_filename, fix_join

        from .tasks import rename_into_caps_task

        write_node = npe.Node(name="WriteCaps", interface=DataSink())
        write_node.inputs.base_directory = str(self.caps_directory)
        write_node.inputs.parameterization = False

        rename_file_node_inputs = [
            "pet_bids_image_filename",
            "pet_preprocessed_image_filename",
            "pet_to_mri_transformation_filename",
            "suvr_reference_region",
            "uncropped_image",
        ]
        if self.parameters.get("save_PETinT1w"):
            rename_file_node_inputs.append("pet_filename_in_t1w_raw")
        rename_files = npe.Node(
            interface=nutil.Function(
                input_names=rename_file_node_inputs,
                output_names=[
                    "pet_filename_caps",
                    "transformation_filename_caps",
                    "pet_filename_in_t1w_caps",
                ],
                function=rename_into_caps_task,
            ),
            name="renameFileCAPS",
        )
        rename_files.interface.inputs.suvr_reference_region = self.parameters.get(
            "suvr_reference_region"
        )
        rename_files.interface.inputs.uncropped_image = self.parameters.get(
            "uncropped_image"
        )
        container_path = npe.Node(
            interface=nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="containerPath",
        )
        self.connect(
            [
                (self.input_node, container_path, [("pet", "bids_or_caps_filename")]),
                (
                    container_path,
                    write_node,
                    [(("container", fix_join, "pet_linear"), "container")],
                ),
                (self.input_node, rename_files, [("pet", "pet_bids_image_filename")]),
                (
                    self.output_node,
                    rename_files,
                    [("affine_mat", "pet_to_mri_transformation_filename")],
                ),
                (
                    rename_files,
                    write_node,
                    [("transformation_filename_caps", "@transform_mat")],
                ),
            ]
        )
        if not (self.parameters.get("uncropped_image")):
            self.connect(
                [
                    (
                        self.output_node,
                        rename_files,
                        [("outfile_crop", "pet_preprocessed_image_filename")],
                    ),
                    (
                        rename_files,
                        write_node,
                        [("pet_filename_caps", "@registered_pet")],
                    ),
                ]
            )
        else:
            self.connect(
                [
                    (
                        self.output_node,
                        rename_files,
                        [("suvr_pet", "pet_preprocessed_image_filename")],
                    ),
                    (
                        rename_files,
                        write_node,
                        [("pet_filename_caps", "@registered_pet")],
                    ),
                ]
            )
        if self.parameters.get("save_PETinT1w"):
            self.connect(
                [
                    (
                        self.output_node,
                        rename_files,
                        [("PETinT1w", "pet_filename_in_t1w_raw")],
                    ),
                    (
                        rename_files,
                        write_node,
                        [("pet_filename_in_t1w_caps", "@registered_pet_in_t1w")],
                    ),
                ]
            )

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces import ants

        from clinica.pipelines.tasks import crop_nifti_task

        from .tasks import perform_suvr_normalization_task
        from .utils import concatenate_transforms, init_input_node, print_end_pipeline

        init_node = npe.Node(
            interface=nutil.Function(
                input_names=["pet"],
                output_names=["pet"],
                function=init_input_node,
            ),
            name="initPipeline",
        )
        concatenate_node = npe.Node(
            interface=nutil.Function(
                input_names=["pet_to_t1w_transform", "t1w_to_mni_transform"],
                output_names=["transforms_list"],
                function=concatenate_transforms,
            ),
            name="concatenateTransforms",
        )

        # The core (processing) nodes

        # 1. `RegistrationSynQuick` by *ANTS*. It uses nipype interface.
        ants_registration_node = npe.Node(
            name="antsRegistration", interface=ants.RegistrationSynQuick()
        )
        ants_registration_node.inputs.dimension = 3
        ants_registration_node.inputs.transform_type = "r"

        # 2. `ApplyTransforms` by *ANTS*. It uses nipype interface. PET to MRI
        ants_applytransform_node = npe.Node(
            name="antsApplyTransformPET2MNI", interface=ants.ApplyTransforms()
        )
        ants_applytransform_node.inputs.dimension = 3
        ants_applytransform_node.inputs.reference_image = self.ref_template

        # 3. Normalize the image (using nifti). It uses custom interface, from utils file
        ants_registration_nonlinear_node = npe.Node(
            name="antsRegistrationT1W2MNI", interface=ants.Registration()
        )
        ants_registration_nonlinear_node.inputs.fixed_image = self.ref_template
        ants_registration_nonlinear_node.inputs.metric = ["MI"]
        ants_registration_nonlinear_node.inputs.metric_weight = [1.0]
        ants_registration_nonlinear_node.inputs.transforms = ["SyN"]
        ants_registration_nonlinear_node.inputs.transform_parameters = [(0.1, 3, 0)]
        ants_registration_nonlinear_node.inputs.dimension = 3
        ants_registration_nonlinear_node.inputs.shrink_factors = [[8, 4, 2]]
        ants_registration_nonlinear_node.inputs.smoothing_sigmas = [[3, 2, 1]]
        ants_registration_nonlinear_node.inputs.sigma_units = ["vox"]
        ants_registration_nonlinear_node.inputs.number_of_iterations = [[200, 50, 10]]
        ants_registration_nonlinear_node.inputs.convergence_threshold = [1e-05]
        ants_registration_nonlinear_node.inputs.convergence_window_size = [10]
        ants_registration_nonlinear_node.inputs.radius_or_number_of_bins = [32]
        ants_registration_nonlinear_node.inputs.winsorize_lower_quantile = 0.005
        ants_registration_nonlinear_node.inputs.winsorize_upper_quantile = 0.995
        ants_registration_nonlinear_node.inputs.collapse_output_transforms = True
        ants_registration_nonlinear_node.inputs.use_histogram_matching = False
        ants_registration_nonlinear_node.inputs.verbose = True

        ants_applytransform_nonlinear_node = npe.Node(
            name="antsApplyTransformNonLinear", interface=ants.ApplyTransforms()
        )
        ants_applytransform_nonlinear_node.inputs.dimension = 3
        ants_applytransform_nonlinear_node.inputs.reference_image = self.ref_template

        if random_seed := self.parameters.get("random_seed", None):
            ants_registration_nonlinear_node.inputs.random_seed = random_seed

        normalize_intensity_node = npe.Node(
            name="intensityNormalization",
            interface=nutil.Function(
                function=perform_suvr_normalization_task,
                input_names=[
                    "pet_image_path",
                    "normalizing_image_path",
                    "reference_mask_path",
                ],
                output_names=["output_img"],
            ),
        )
        normalize_intensity_node.inputs.reference_mask_path = self.ref_mask

        # 4. Crop image (using nifti). It uses custom interface, from utils file
        crop_nifti_node = npe.Node(
            name="cropNifti",
            interface=nutil.Function(
                function=crop_nifti_task,
                input_names=["input_image", "output_path"],
                output_names=["output_image"],
            ),
        )
        crop_nifti_node.inputs.output_path = self.base_dir

        # 5. Print end message
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["pet", "final_file"], function=print_end_pipeline
            ),
            name="WriteEndMessage",
        )

        # 6. Optional node: compute PET image in T1w
        ants_applytransform_optional_node = npe.Node(
            name="antsApplyTransformPET2T1w", interface=ants.ApplyTransforms()
        )
        ants_applytransform_optional_node.inputs.dimension = 3

        self.connect(
            [
                (self.input_node, init_node, [("pet", "pet")]),
                # STEP 1
                (self.input_node, ants_registration_node, [("t1w", "fixed_image")]),
                (init_node, ants_registration_node, [("pet", "moving_image")]),
                # STEP 2
                (
                    ants_registration_node,
                    concatenate_node,
                    [("out_matrix", "pet_to_t1w_transform")],
                ),
                (
                    self.input_node,
                    concatenate_node,
                    [("t1w_to_mni", "t1w_to_mni_transform")],
                ),
                (self.input_node, ants_applytransform_node, [("pet", "input_image")]),
                (
                    concatenate_node,
                    ants_applytransform_node,
                    [("transforms_list", "transforms")],
                ),
                # STEP 3
                (
                    self.input_node,
                    ants_registration_nonlinear_node,
                    [("t1w", "moving_image")],
                ),
                (
                    ants_registration_nonlinear_node,
                    ants_applytransform_nonlinear_node,
                    [("reverse_forward_transforms", "transforms")],
                ),
                (
                    ants_applytransform_node,
                    ants_applytransform_nonlinear_node,
                    [("output_image", "input_image")],
                ),
                (
                    ants_applytransform_node,
                    normalize_intensity_node,
                    [("output_image", "pet_image_path")],
                ),
                (
                    ants_applytransform_nonlinear_node,
                    normalize_intensity_node,
                    [("output_image", "normalizing_image_path")],
                ),
                # Connect to DataSink
                (
                    ants_registration_node,
                    self.output_node,
                    [("out_matrix", "affine_mat")],
                ),
                (
                    normalize_intensity_node,
                    self.output_node,
                    [("output_img", "suvr_pet")],
                ),
                (self.input_node, print_end_message, [("pet", "pet")]),
            ]
        )
        # STEP 4
        if not (self.parameters.get("uncropped_image")):
            self.connect(
                [
                    (
                        normalize_intensity_node,
                        crop_nifti_node,
                        [("output_img", "input_image")],
                    ),
                    (
                        crop_nifti_node,
                        self.output_node,
                        [("output_image", "outfile_crop")],
                    ),
                    (
                        crop_nifti_node,
                        print_end_message,
                        [("output_image", "final_file")],
                    ),
                ]
            )
        else:
            self.connect(
                [
                    (
                        normalize_intensity_node,
                        print_end_message,
                        [("output_img", "final_file")],
                    ),
                ]
            )
        # STEP 6: Optional argument
        if self.parameters.get("save_PETinT1w"):
            self.connect(
                [
                    (
                        self.input_node,
                        ants_applytransform_optional_node,
                        [("pet", "input_image"), ("t1w", "reference_image")],
                    ),
                    (
                        ants_registration_node,
                        ants_applytransform_optional_node,
                        [("out_matrix", "transforms")],
                    ),
                    (
                        ants_applytransform_optional_node,
                        self.output_node,
                        [("output_image", "PETinT1w")],
                    ),
                ]
            )



import xml.etree.ElementTree
from pathlib import Path
from typing import Iterable, Optional, Tuple, Union

import pandas as pd

from clinica.utils.exceptions import ClinicaXMLParserError
from clinica.utils.stream import cprint, log_and_raise

__all__ = ["create_json_metadata"]


LOGGING_HEADER = "[ADNI JSON]"


def _log_and_raise_xml_parser_error(message: str):
    log_and_raise(f"{LOGGING_HEADER} {message}", ClinicaXMLParserError)


# Map from names of extracted metadata to their proper BIDS names.
METADATA_NAME_MAPPING = {
    "acquisition_type": "MRAcquisitionType",
    "pulse_sequence": "PulseSequenceType",
    "manufacturer": "Manufacturer",
    "field_strength": "MagneticFieldStrength",
}


def _bids_id_to_loni(bids_id: str) -> Optional[str]:
    """Convert a subject id of the form sub-ADNI000S0000
    back to original format 000_S_0000
    """
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    try:
        return bids_id_factory(StudyName.ADNI)(bids_id).to_original_study_id()
    except ValueError:
        return None


def _read_xml_files(
    subject_ids: Optional[Iterable[str]] = None,
    xml_path: Optional[Path] = None,
) -> Iterable[Path]:
    """Return the XML files in the folder `xml_path` for the provided `subj_ids`.
    This function assumes that file are named "ADNI_{sub_ids}.xml".
    If no files were found, an `IndexError` is raised.
    """
    xml_path = xml_path or Path.cwd()
    if subject_ids:
        xml_files = []
        xml_regex = [f"ADNI_{e}*.xml" for e in subject_ids]
        for subj_files in xml_regex:
            xml_files.extend([f for f in xml_path.glob(subj_files)])
    else:
        xml_files = [f for f in xml_path.glob("Clinica_processed_metadata/ADNI_*.xml")]
    if len(xml_files) == 0:
        msg = f"{LOGGING_HEADER} No ADNI xml files were found for reading the metadata."
        cprint(msg, lvl="error")
        raise FileNotFoundError(msg)
    return xml_files


def _check_xml_tag(element_tag: str, expected_tag: str):
    """Check that the XML element tagged matches the expected tag."""
    if element_tag != expected_tag:
        _log_and_raise_xml_parser_error(
            f"Bad tag: expected {expected_tag}, got {element_tag}"
        )


def _check_xml_nb_children(
    xml_el: xml.etree.ElementTree.Element,
    expected_nb_children: Union[int, Iterable[int]],
):
    """Check that the given XML element has the expected number of children.
    Raise a `ValueError` otherwise.
    """
    nb_children = len(xml_el)
    if isinstance(expected_nb_children, int):
        if nb_children != expected_nb_children:
            _log_and_raise_xml_parser_error(
                f"Bad number of children for <{xml_el.tag}>: "
                f"got {nb_children} != {expected_nb_children}"
            )
    else:
        if nb_children not in expected_nb_children:
            _log_and_raise_xml_parser_error(
                f"Bad number of children for <{xml_el.tag}>: "
                f"got {nb_children}, not in {expected_nb_children}"
            )


def _check_xml(
    xml_el: xml.etree.ElementTree.Element,
    exp_tag: str,
    exp_nb_children: Optional[Union[int, Iterable[int]]] = None,
) -> xml.etree.ElementTree.Element:
    """Check that the given xml element is as expected (tag name + nb of children).
    Raise a ValueError if not and return the element if it is valid.
    """
    _check_xml_tag(xml_el.tag, exp_tag)
    if exp_nb_children is not None:
        _check_xml_nb_children(xml_el, exp_nb_children)
    return xml_el


def _parse_project(
    root: xml.etree.ElementTree.Element,
) -> xml.etree.ElementTree.Element:
    """Check that root has 1 child: <project> having 4 children,
    extract project from them.
    Check that the project identifier contains ADNI.
    """
    if len(root) != 1:
        _log_and_raise_xml_parser_error(
            "XML root should have only one child. " f"{len(root)} children found.",
        )
    project = _check_xml(root[0], "project", 4)
    _check_xml_project_identifier(project, "ADNI")
    return project


def _check_xml_and_get_text(
    xml_element: xml.etree.ElementTree.Element, expected_tag: str, *, cast=None
) -> str:
    """Check xml element and return its text (leaf)"""
    _check_xml(xml_element, expected_tag, 0)  # leaf
    return _get_text(xml_element, cast=cast)


def _get_text(xml_element: xml.etree.ElementTree.Element, cast=None) -> str:
    """Get text of xml element (leaf)."""
    if cast is not None:
        return cast(xml_element.text)
    return xml_element.text


def _check_derived_image_xml(derived: xml.etree.ElementTree.Element) -> None:
    """Perform sanity checks on derived images."""
    if len(derived) < 9:
        _log_and_raise_xml_parser_error("Derived image does not have enough field.")
    for idx, field, expected in zip(
        [2, 3, 4, 5, 6],
        ["imageType", "tissue", "hemisphere", "anatomicStructure", "registration"],
        ["image volume", "All", "Both", "Brain", "native"],
    ):
        _check_xml_field(derived[idx], field, expected)
    idx = -1 if derived[-1].tag == "creationDate" else -2
    if (
        image_create_date := _check_xml_and_get_text(derived[idx], "creationDate")
    ) != "0000-00-00":
        try:
            _validate_date_iso_format(image_create_date)
        except ValueError as e:
            _log_and_raise_xml_parser_error(
                f"The creationDate for the derived image is not valid: {e}"
            )


def _check_xml_field(
    xml_element: xml.etree.ElementTree.Element, field: str, expected_value: str
) -> None:
    """Check that the given field of the given XML element has the expected value."""
    if (found_value := _check_xml_and_get_text(xml_element, field)) != expected_value:
        _log_and_raise_xml_parser_error(
            f"The {field} for the derived image should be '{expected_value}'. "
            f"{found_value} was found instead."
        )


def _validate_date_iso_format(date: str) -> None:
    import datetime

    try:
        datetime.date.fromisoformat(date)
    except ValueError:
        raise ValueError(f"Incorrect data format {date}, should be YYYY-MM-DD")


def _get_derived_image_metadata(derived: xml.etree.ElementTree.Element) -> dict:
    """Return metadata for derived images after
    performing some sanity checks.
    """
    _check_derived_image_xml(derived)
    return {
        "image_proc_id": _check_xml_and_get_text(derived[0], "imageUID", cast=int),
        "image_proc_desc": _check_xml_and_get_text(
            derived[1], "processedDataLabel"
        ).strip(),
    }


def _check_xml_project_identifier(
    project: xml.etree.ElementTree.Element, expected: str
):
    """Check the project identifier."""
    if _check_xml_and_get_text(project[0], "projectIdentifier") != expected:
        _log_and_raise_xml_parser_error(f"Not {expected} cohort")


def _get_original_image_metadata(original_image: xml.etree.ElementTree.Element) -> dict:
    """Get original image metadata."""
    if not isinstance(original_image, list):
        return _get_image_metadata(original_image)
    # only keep common metadata (remaining will be None) and
    original_images_metadata = [_get_image_metadata(_) for _ in original_image]
    original_images_metadata = {
        _.pop("image_orig_id"): _ for _ in original_images_metadata
    }
    original_image_metadata = {
        "image_orig_id": "|".join(map(str, original_images_metadata.keys()))
    }
    for k in next(iter(original_images_metadata.values())).keys():
        # loop on original images metadata and only add consistent metadata
        set_v = set(_[k] for _ in original_images_metadata.values())
        if len(set_v) == 1:
            original_image_metadata[k] = set_v.pop()
    return original_image_metadata


def _check_processed_image_rating(
    processed_image_rating: Union[int, None],
    original_image_metadata: dict,
    derived_image_metadata: dict,
):
    """Check, if possible, that the rating of the processed images are
    consistent with the rating of the original image.
    """
    from clinica.utils.stream import cprint

    rating = original_image_metadata.get("image_orig_rating", None)
    if processed_image_rating is not None:
        if processed_image_rating != rating:
            cprint(
                msg=(
                    f"{LOGGING_HEADER} Image rating for processed image "
                    f"{derived_image_metadata.get('image_proc_id')} not "
                    "consistent with rating of original image"
                ),
                lvl="info",
            )


def _get_root_from_xml_path(xml_path: Path) -> xml.etree.ElementTree.Element:
    """Return the root XML element from the XML file path."""
    from xml.etree import ElementTree

    try:
        tree = ElementTree.parse(xml_path)
        root = tree.getroot()
        return root
    except Exception as e:
        _log_and_raise_xml_parser_error(
            f"Error parsing XML file {xml_path.name} :\n{e}"
        )


def _parse_series(
    study: xml.etree.ElementTree.Element,
) -> xml.etree.ElementTree.Element:
    """Get series xml element from study xml element."""
    expected_length = 3 if len(study) == 7 else 4
    return _check_xml(study[5], "series", expected_length)


def _parse_derived_images(
    series: xml.etree.ElementTree.Element,
) -> Tuple[Union[xml.etree.ElementTree.Element, list], Union[int, None], dict]:
    """Parse the series xml element when derived images are present.
    Return the original image, the rating of the processed images,
    and the derived image metadata as a dict.
    """
    # we can have 1 or 2 related imgs...
    series_meta = _check_xml(series[-1], "seriesLevelMeta", {3, 4})
    related_image = _check_xml(series_meta[2], "relatedImageDetail", 1)
    original_image = _check_xml(related_image[0], "originalRelatedImage", {3, 4})
    if len(series_meta) == 4:
        # 2 original images
        related_image_bis = _check_xml(series_meta[-1], "relatedImageDetail", 1)
        original_image_bis = _check_xml(
            related_image_bis[0], "originalRelatedImage", len(original_image)
        )
        original_image = [original_image, original_image_bis]
    # specific
    pipeline_name = _check_xml_and_get_text(
        _check_xml(series_meta[0], "annotation", 1)[0], "text"
    )
    deriv = _check_xml(series_meta[1], "derivedProduct")
    derived_metadata = {"image_proc_pipe": pipeline_name.strip()}
    derived_metadata.update(_get_derived_image_metadata(deriv))
    processed_image_rating = _get_image_rating(deriv[-1])
    return original_image, processed_image_rating, derived_metadata


def _check_image(img: xml.etree.ElementTree.Element):
    """Check that the image XML element is valid."""
    if img.tag not in {
        "imagingProtocol",  # for original image metadata
        "originalRelatedImage",  # for processed image
    }:
        _log_and_raise_xml_parser_error(
            f"Bad image tag <{img.tag}>. "
            "Should be either 'imagingProtocol' or 'originalRelatedImage'."
        )
    if len(img) not in {3, 4}:
        _log_and_raise_xml_parser_error(
            f"Image XML element has {len(img)} children. " "Expected either 3 or 4.",
        )


def _clean_protocol_metadata(protocol_metadata: dict) -> dict:
    """Replace confusing '|' (reserved for separation btw elements) with space.
    Only 1 manufacturer with this apparently but...
    """
    return {
        k: v.replace("|", " ")
        if isinstance(v, str)
        else v  # there are some None (not str)
        for k, v in protocol_metadata.items()
    }


def _filter_metadata(metadata: dict) -> dict:
    """Filter and clean a given metadata dictionary according to the
    METADATA_NAME_MAPPING dictionary."""
    filtered = dict()
    for k, v in metadata.items():
        if k in METADATA_NAME_MAPPING:
            filtered[METADATA_NAME_MAPPING[k]] = v
    return filtered


def _get_image_metadata(img: xml.etree.ElementTree.Element) -> dict:
    """Return information on original image as dict of metadata."""
    _check_image(img)
    protocol = _check_xml(img[2], "protocolTerm")  # do not check children
    assert all(p.tag == "protocol" and "term" in p.attrib.keys() for p in protocol)
    protocol_metadata = {
        "_".join(p.attrib["term"].split()).lower(): p.text for p in protocol
    }
    protocol_metadata = _clean_protocol_metadata(protocol_metadata)
    filtered_protocol_metadata = _filter_metadata(protocol_metadata)
    return {
        "image_orig_id": _check_xml_and_get_text(img[0], "imageUID", cast=int),
        "image_orig_seq": _check_xml_and_get_text(
            img[1], "description"
        ),  # .replace('|', ' '),
        **filtered_protocol_metadata,
    }


def _get_image_rating(image_rating: xml.etree.ElementTree.Element) -> Optional[str]:
    """Get the image rating value as an integer from the xml element."""
    if image_rating.tag != "imageRating":
        return None
    _check_xml(image_rating, "imageRating", 2)
    image_rating_desc = _check_xml_and_get_text(image_rating[0], "ratingDescription")
    image_rating_val = _check_xml_and_get_text(image_rating[1], "value", cast=int)
    assert image_rating_desc == str(image_rating_val)
    return image_rating_val


def _check_modality(study: xml.etree.ElementTree.Element, expected_modality: str):
    """Check that the modality of the given study is the expected one."""
    series = _parse_series(study)
    modality = _check_xml_and_get_text(series[1], "modality")
    if modality != expected_modality:
        _log_and_raise_xml_parser_error(
            f"Unexpected modality {modality}, expected {expected_modality}."
        )


def _parse_subject(
    project: xml.etree.ElementTree.Element,
) -> Tuple[str, xml.etree.ElementTree.Element]:
    """From the project xml element, parse the subject and subject id."""
    subject = _check_xml(project[-1], "subject", {5, 7})  # with APOE or not
    subject_id = _check_xml_and_get_text(subject[0], "subjectIdentifier")
    return subject_id, subject


def _parse_study(
    subject: xml.etree.ElementTree.Element,
) -> xml.etree.ElementTree.Element:
    """Parse the study from the subject xml element
    and check that it is MRI.
    """
    study = _check_xml(subject[-1], "study", {6, 7})
    _check_modality(study, "MRI")
    return study


def _parse_images(
    study: xml.etree.ElementTree.Element,
) -> Tuple[Union[xml.etree.ElementTree.Element, list], Union[int, None], dict]:
    """Parse the images, original and derived if available."""
    if len(study) == 7:
        # Original only
        original_image = _check_xml(study[6], "imagingProtocol", {3, 4})
        return original_image, None, dict()
    else:
        series = _parse_series(study)
        return _parse_derived_images(series)


def _parse_xml_file(xml_path: Path) -> dict:
    """Parse the given XML file and return the desired metadata as a dict."""
    project = _parse_project(_get_root_from_xml_path(xml_path))
    subject_id, subject = _parse_subject(project)
    study = _parse_study(subject)
    series = _parse_series(study)
    original_image, processed_image_rating, derived_image_metadata = _parse_images(
        study
    )
    original_image_metadata = _get_original_image_metadata(original_image)
    _check_processed_image_rating(
        processed_image_rating, original_image_metadata, derived_image_metadata
    )
    acquisition_date = pd.Timestamp(_check_xml_and_get_text(series[2], "dateAcquired"))
    scan_metadata = {
        "id": subject_id,
        "acq_time": acquisition_date.isoformat(),
        **original_image_metadata,
        **derived_image_metadata,
    }
    if "image_proc_id" not in scan_metadata and "image_orig_id" not in scan_metadata:
        _log_and_raise_xml_parser_error(
            f"Scan metadata for subject {subject_id} has no image ID."
        )
    return scan_metadata


class FuncWithException:
    def __init__(self, f):
        self.f = f

    def __call__(self, *args, **kwargs):
        """Return a tuple: (result, exception)"""
        try:
            return self.f(*args, **kwargs), None
        except Exception as e:
            return None, e


def _run_parsers(xml_files: Iterable[Path]) -> Tuple[Iterable[dict], dict]:
    """Run the parser `_parse_xml_file` on the list of files `xml_files`.
    Returns a tuple consisting if parsed images
    metadata and captured exceptions.
    """
    import os

    parser = FuncWithException(_parse_xml_file)
    images_with_exceptions = dict(
        zip(
            xml_files,  # tqdm is buggy when chunksize > 1
            [parser(xml_file) for xml_file in xml_files],
        )
    )
    images = [
        image
        for image, exception in images_with_exceptions.values()
        if exception is None
    ]
    exceptions = {
        os.path.basename(xml_p): exception
        for xml_p, (_, exception) in images_with_exceptions.items()
        if exception is not None
    }
    return images, exceptions


def _create_mri_meta_df(images: Iterable[dict]) -> pd.DataFrame:
    """Create dataframe from images metadata parsed from the XML files."""
    mri_meta = pd.DataFrame(images).convert_dtypes()
    # Scan ID as expected by Clinica (derived if so, original otherwise)
    mri_meta["T1w_scan_id"] = mri_meta["image_proc_id"].combine_first(
        mri_meta["image_orig_id"]
    )
    # Index
    mri_meta.set_index(
        [
            "T1w_scan_id",
            "id",
            # "GENDER",
            # "VISIT_DESC",
            # "VISIT_AGE",
            # "SCAN_DATE",
            # "STUDY_ID",
            # "SERIES_ID",
        ],
        inplace=True,
    )
    mri_meta.reset_index(inplace=True)
    return mri_meta


def _get_existing_scan_dataframe(
    subject_path: Path, session: str
) -> Tuple[Optional[pd.DataFrame], Path]:
    """Retrieve existing scan dataframe at the given
    `subj_path`, and the given `session`.
    If no existing scan file is found, a warning is given.
    """
    import warnings

    from clinica.utils.stream import cprint

    scans_tsv_path = subject_path / session / f"{subject_path.name}_{session}_scans.tsv"
    if scans_tsv_path.exists():
        df_scans = pd.read_csv(scans_tsv_path, sep="\t")
        df_scans["scan_id"] = df_scans["scan_id"].astype("Int64")
        return df_scans, scans_tsv_path
    msg = f"{LOGGING_HEADER} No scan tsv file for subject {subject_path.name} and session {session}"
    cprint(msg, lvl="warning")
    warnings.warn(msg)
    return None, scans_tsv_path


def _merge_scan_and_metadata(
    df_scans: pd.DataFrame, df_meta: pd.DataFrame, strategy: dict
) -> pd.DataFrame:
    """Perform a merge between the two provided dataframe
    according to the strategy.
    """
    return pd.merge(df_scans, df_meta, **strategy)


def _get_json_filename_from_scan_filename(scan_filename: Path) -> Path:
    """Replace the given `scan_filename` extensions with '.json'."""
    extensions = "".join(scan_filename.suffixes)
    return Path(str(scan_filename).replace(extensions, ".json"))


def _add_json_scan_metadata(
    json_path: Path,
    metadata: Union[dict, str],
    indent: int = 4,
    keep_none: bool = False,
):
    """Load existing metadata at the provided `json_path`, merge with new
    provided `metadata`, and finally write to disk at `json_path`.

    .. warning::
        This overwrites the existing json file.
    """
    import json

    existing_metadata = dict()
    if json_path.exists():
        with open(json_path, "r") as fp:
            existing_metadata = json.load(fp)
    if isinstance(metadata, str):
        metadata = json.loads(metadata)
    filtered_metadata = {
        k: v for k, v in metadata.items() if k in METADATA_NAME_MAPPING.values()
    }
    updated_metadata = {**existing_metadata, **filtered_metadata}
    if not keep_none:
        updated_metadata = {k: v for k, v in updated_metadata.items() if v is not None}
    if len(updated_metadata) > 0:
        with open(json_path, "w") as fp:
            json.dump(updated_metadata, fp, indent=indent)


def _add_metadata_to_scans(df_meta: pd.DataFrame, bids_subjects_paths: Iterable[Path]):
    """Add the metadata to the appropriate tsv and json files."""
    from clinica.iotools.bids_utils import get_bids_sess_list

    merge_strategy = {"how": "left", "left_on": "scan_id", "right_on": "T1w_scan_id"}

    for subject_path in bids_subjects_paths:
        if (sess_list := get_bids_sess_list(subject_path)) is not None:
            for sess in sess_list:
                df_scans, scans_tsv_path = _get_existing_scan_dataframe(
                    subject_path, sess
                )
                if df_scans is not None:
                    columns_to_keep = list(df_scans.columns) + ["acq_time"]
                    df_merged = _merge_scan_and_metadata(
                        df_scans, df_meta, merge_strategy
                    )
                    for _, scan_row in df_merged.iterrows():
                        _add_json_scan_metadata(
                            _get_json_filename_from_scan_filename(
                                subject_path / sess / scan_row["filename"]
                            ),
                            scan_row.to_json(),
                        )
                    df_merged[columns_to_keep].to_csv(
                        scans_tsv_path, sep="\t", index=False
                    )


def create_json_metadata(
    bids_subjects_paths: Iterable[Path], bids_ids: Iterable[str], xml_path: Path
):
    """Create json metadata dictionary and add the metadata to the
    appropriate files in the BIDS hierarchy."""
    loni_ids = [_bids_id_to_loni(bids_id) for bids_id in bids_ids]
    xml_files = _read_xml_files(loni_ids, xml_path)
    images, exe = _run_parsers(xml_files)
    df_meta = _create_mri_meta_df(images)
    _add_metadata_to_scans(df_meta, bids_subjects_paths)
